import unittest
import base64
import helics as h

from dots_infrastructure.EsdlHelper import EsdlHelper
from dots_infrastructure.DataClasses import CalculationServiceInput, SubscriptionDescription

class TestParse(unittest.TestCase):

    def setUp(self):
        with open("test-input-extraction-network.esdl", mode="r") as esdl_file:
            self.encoded_base64_esdl = base64.b64encode(esdl_file.read().encode('utf-8')).decode('utf-8')

    def test_assets_in_building_can_get_inputs_from_other_assets_inside_building(self):
        # Arrange
        simulator_esdl_id = '15bc27e8-97db-427c-959e-e2a2fca27f75'

        esdl_helper = EsdlHelper(self.encoded_base64_esdl)

        subscription_descriptions = [
            SubscriptionDescription(esdl_type="ElectricityDemand",input_name="active_power",input_unit="W",input_type=h.HelicsDataType.VECTOR),
            SubscriptionDescription(esdl_type="ElectricityDemand",input_name="reactive_power",input_unit="VAr",input_type=h.HelicsDataType.VECTOR),
            SubscriptionDescription(esdl_type="EConnection",input_name="heat_to_dw",input_unit="W",input_type=h.HelicsDataType.VECTOR),
            SubscriptionDescription(esdl_type="HybridHeatPump",input_name="buffer_temperature",input_unit="K",input_type=h.HelicsDataType.DOUBLE),
            SubscriptionDescription(esdl_type="HybridHeatPump",input_name="house_temperatures",input_unit="K",input_type=h.HelicsDataType.VECTOR),
            SubscriptionDescription(esdl_type="EVChargingStation",input_name="state_of_charge_ev",input_unit="J",input_type=h.HelicsDataType.DOUBLE)
        ]

        expected_input_descriptions = [
            CalculationServiceInput("EConnection", "heat_to_dw", '7415cddb-b735-4646-b772-47f101b5c7a8', "W", h.HelicsDataType.VECTOR, simulator_esdl_id, "EConnection/heat_to_dw/7415cddb-b735-4646-b772-47f101b5c7a8"),
            CalculationServiceInput("ElectricityDemand", "active_power", '5ad97622-7226-40b1-a163-260b3478b1e3', "W", h.HelicsDataType.VECTOR, simulator_esdl_id, "ElectricityDemand/active_power/5ad97622-7226-40b1-a163-260b3478b1e3"),
            CalculationServiceInput("ElectricityDemand", "reactive_power", '5ad97622-7226-40b1-a163-260b3478b1e3', "VAr", h.HelicsDataType.VECTOR, simulator_esdl_id, "ElectricityDemand/reactive_power/5ad97622-7226-40b1-a163-260b3478b1e3")
        ]

        calculation_services = [
            "ElectricityDemand",
            "EConnection"
        ]

        # Execute
        inputs = esdl_helper.get_connected_input_esdl_objects(simulator_esdl_id, calculation_services, subscription_descriptions)

        # Assert correct assets are extracted from esdl file
        self.assertListEqual(expected_input_descriptions, inputs)
    
    def test_esdl_entity_recevies_subscriptions_from_connected_entities(self):

        # Arrange
        simulator_esdl_id = '7415cddb-b735-4646-b772-47f101b5c7a8'

        esdl_helper = EsdlHelper(self.encoded_base64_esdl)

        subscription_descriptions = [
            SubscriptionDescription(esdl_type="ElectricityDemand",input_name="active_power",input_unit="W",input_type=h.HelicsDataType.VECTOR),
            SubscriptionDescription(esdl_type="ElectricityDemand",input_name="reactive_power",input_unit="VAr",input_type=h.HelicsDataType.VECTOR),
            SubscriptionDescription(esdl_type="PVInstallation",input_name="potential_active_power",input_unit="W",input_type=h.HelicsDataType.VECTOR),
            SubscriptionDescription(esdl_type="HybridHeatPump",input_name="buffer_temperature",input_unit="K",input_type=h.HelicsDataType.DOUBLE),
            SubscriptionDescription(esdl_type="HybridHeatPump",input_name="house_temperatures",input_unit="K",input_type=h.HelicsDataType.VECTOR),
            SubscriptionDescription(esdl_type="EVChargingStation",input_name="state_of_charge_ev",input_unit="J",input_type=h.HelicsDataType.DOUBLE)
        ]

        expected_input_descriptions = [
            CalculationServiceInput("ElectricityDemand", "active_power", '5ad97622-7226-40b1-a163-260b3478b1e3', "W", h.HelicsDataType.VECTOR, simulator_esdl_id, "ElectricityDemand/active_power/5ad97622-7226-40b1-a163-260b3478b1e3"),
            CalculationServiceInput("ElectricityDemand", "reactive_power", '5ad97622-7226-40b1-a163-260b3478b1e3', "VAr", h.HelicsDataType.VECTOR, simulator_esdl_id, "ElectricityDemand/reactive_power/5ad97622-7226-40b1-a163-260b3478b1e3"),
            CalculationServiceInput("PVInstallation", "potential_active_power", '208c4a92-148a-4893-b474-37cad47b2fcb', "W", h.HelicsDataType.VECTOR, simulator_esdl_id, "PVInstallation/potential_active_power/208c4a92-148a-4893-b474-37cad47b2fcb"),
            CalculationServiceInput("EVChargingStation", "state_of_charge_ev", '2c285e74-f018-4305-bdf5-dd0f49fcbeab', "J", h.HelicsDataType.DOUBLE, simulator_esdl_id, "EVChargingStation/state_of_charge_ev/2c285e74-f018-4305-bdf5-dd0f49fcbeab"),
            CalculationServiceInput("HybridHeatPump", "buffer_temperature", '15bc27e8-97db-427c-959e-e2a2fca27f75', "K", h.HelicsDataType.DOUBLE, simulator_esdl_id, "HybridHeatPump/buffer_temperature/15bc27e8-97db-427c-959e-e2a2fca27f75"),
            CalculationServiceInput("HybridHeatPump", "house_temperatures", '15bc27e8-97db-427c-959e-e2a2fca27f75', "K", h.HelicsDataType.VECTOR, simulator_esdl_id, "HybridHeatPump/house_temperatures/15bc27e8-97db-427c-959e-e2a2fca27f75")
        ]

        calculation_services = [
            "ElectricityDemand",
            "PVInstallation",
            "EVChargingStation",
            "HeatPump",
            "HybridHeatPump"
        ]

        # Execute
        inputs = esdl_helper.get_connected_input_esdl_objects(simulator_esdl_id, calculation_services, subscription_descriptions)

        # Assert correct assets are extracted from esdl file
        self.assertListEqual(expected_input_descriptions, inputs)

    def test_esdl_entity_can_subscribe_to_non_connected_inputs(self):
        simulator_esdl_id = '7415cddb-b735-4646-b772-47f101b5c7a8'

        esdl_helper = EsdlHelper(self.encoded_base64_esdl)

        subscription_descriptions = [
            SubscriptionDescription("EnergyMarket", "DA_Price", "EUR", h.HelicsDataType.DOUBLE)
        ]

        expected_input_descriptions = [
            CalculationServiceInput("EnergyMarket", "DA_Price", '80f75d42-80a8-446e-8611-cb24154f2bd5', "EUR", h.HelicsDataType.DOUBLE, simulator_esdl_id, "EnergyMarket/DA_Price/80f75d42-80a8-446e-8611-cb24154f2bd5")
        ]

        calculation_services = [
            "EnergyMarket"
        ]

        # Execute
        inputs = esdl_helper.get_connected_input_esdl_objects(simulator_esdl_id, calculation_services, subscription_descriptions)

        # Assert correct assets are extracted from esdl file
        self.assertListEqual(expected_input_descriptions, inputs)

    def test_esdl_entity_has_two_publications_then_connected_entity_can_subscribe_to_both(self):

        # Arrange

        simulator_esdl_id = '7415cddb-b735-4646-b772-47f101b5c7a8'

        esdl_helper = EsdlHelper(self.encoded_base64_esdl)

        subscription_descriptions = [
            SubscriptionDescription("PVInstallation", "PV_Dispatch", "W", h.HelicsDataType.DOUBLE),
            SubscriptionDescription("PVInstallation", "PV_Dispatch2", "W", h.HelicsDataType.DOUBLE)
        ]

        expected_input_descriptions = [
            CalculationServiceInput("PVInstallation", "PV_Dispatch", '208c4a92-148a-4893-b474-37cad47b2fcb', "W", h.HelicsDataType.DOUBLE, simulator_esdl_id, "PVInstallation/PV_Dispatch/208c4a92-148a-4893-b474-37cad47b2fcb"),
            CalculationServiceInput("PVInstallation", "PV_Dispatch2", '208c4a92-148a-4893-b474-37cad47b2fcb', "W", h.HelicsDataType.DOUBLE, simulator_esdl_id, "PVInstallation/PV_Dispatch2/208c4a92-148a-4893-b474-37cad47b2fcb")
        ]

        calculation_services = [
            "PVInstallation"
        ]

        # Execute
        inputs = esdl_helper.get_connected_input_esdl_objects(simulator_esdl_id, calculation_services, subscription_descriptions)

        # Assert correct assets are extracted from esdl file
        self.assertListEqual(expected_input_descriptions, inputs)

    def test_non_energy_entity_subscriptions_are_correctly_extracted(self):
        # Arrange
        simulator_esdl_id = '06c59a5e-aa84-4a4d-90db-56fbe4eb266c'

        esdl_helper = EsdlHelper(self.encoded_base64_esdl)

        subscription_descriptions = [
            SubscriptionDescription("EConnection", "EConnectionDispatch", "W", h.HelicsDataType.DOUBLE)
        ]

        expected_input_descriptions = [
            CalculationServiceInput("EConnection", "EConnectionDispatch", '7415cddb-b735-4646-b772-47f101b5c7a8', "W", h.HelicsDataType.DOUBLE, simulator_esdl_id, "EConnection/EConnectionDispatch/7415cddb-b735-4646-b772-47f101b5c7a8"),
            CalculationServiceInput("EConnection", "EConnectionDispatch", 'fd7fc047-30b1-48e3-99d9-1bc882772170', "W", h.HelicsDataType.DOUBLE, simulator_esdl_id, "EConnection/EConnectionDispatch/fd7fc047-30b1-48e3-99d9-1bc882772170"),
            CalculationServiceInput("EConnection", "EConnectionDispatch", '8f155685-c164-4d0c-ad3f-3419f4b97c82', "W", h.HelicsDataType.DOUBLE, simulator_esdl_id, "EConnection/EConnectionDispatch/8f155685-c164-4d0c-ad3f-3419f4b97c82"),
            CalculationServiceInput("EConnection", "EConnectionDispatch", '1f776d2a-ea5b-43e1-94a1-aac7e8381e41', "W", h.HelicsDataType.DOUBLE, simulator_esdl_id, "EConnection/EConnectionDispatch/1f776d2a-ea5b-43e1-94a1-aac7e8381e41"),
            CalculationServiceInput("EConnection", "EConnectionDispatch", '8e43e731-e4a4-46ac-a5b0-7aec5031878b', "W", h.HelicsDataType.DOUBLE, simulator_esdl_id, "EConnection/EConnectionDispatch/8e43e731-e4a4-46ac-a5b0-7aec5031878b"),
            CalculationServiceInput("EConnection", "EConnectionDispatch", 'f10ff819-1d81-4cc7-82dc-b02ffbc60bfc', "W", h.HelicsDataType.DOUBLE, simulator_esdl_id, "EConnection/EConnectionDispatch/f10ff819-1d81-4cc7-82dc-b02ffbc60bfc"),
            CalculationServiceInput("EConnection", "EConnectionDispatch", '221c77de-74f8-4293-bb4e-2dd39dee1acd', "W", h.HelicsDataType.DOUBLE, simulator_esdl_id, "EConnection/EConnectionDispatch/221c77de-74f8-4293-bb4e-2dd39dee1acd"),
            CalculationServiceInput("EConnection", "EConnectionDispatch", '18f85d6d-6051-4cdf-ab4c-e80a404e3c74', "W", h.HelicsDataType.DOUBLE, simulator_esdl_id, "EConnection/EConnectionDispatch/18f85d6d-6051-4cdf-ab4c-e80a404e3c74"),
            CalculationServiceInput("EConnection", "EConnectionDispatch", '891c0abb-f6e8-4eb3-9045-d6c87120d281', "W", h.HelicsDataType.DOUBLE, simulator_esdl_id, "EConnection/EConnectionDispatch/891c0abb-f6e8-4eb3-9045-d6c87120d281"),
            CalculationServiceInput("EConnection", "EConnectionDispatch", 'a6c6d5d0-ca19-4579-85f9-8fd99f77aaf3', "W", h.HelicsDataType.DOUBLE, simulator_esdl_id, "EConnection/EConnectionDispatch/a6c6d5d0-ca19-4579-85f9-8fd99f77aaf3"),
            CalculationServiceInput("EConnection", "EConnectionDispatch", 'e69bae01-d2db-4829-bbb7-71148ead969b', "W", h.HelicsDataType.DOUBLE, simulator_esdl_id, "EConnection/EConnectionDispatch/e69bae01-d2db-4829-bbb7-71148ead969b"),
            CalculationServiceInput("EConnection", "EConnectionDispatch", 'c569d7a1-d2d7-4c05-8e31-c641ea48e9ff', "W", h.HelicsDataType.DOUBLE, simulator_esdl_id, "EConnection/EConnectionDispatch/c569d7a1-d2d7-4c05-8e31-c641ea48e9ff"),
            CalculationServiceInput("EConnection", "EConnectionDispatch", 'b2c3bb3b-0b40-4bc2-975c-cd648fe03065', "W", h.HelicsDataType.DOUBLE, simulator_esdl_id, "EConnection/EConnectionDispatch/b2c3bb3b-0b40-4bc2-975c-cd648fe03065")
        ]

        calculation_services = [
            "EConnection"
        ]

        # Execute
        inputs = esdl_helper.get_connected_input_esdl_objects(simulator_esdl_id, calculation_services, subscription_descriptions)

        # Assert correct assets are extracted from esdl file
        self.assertListEqual(expected_input_descriptions, inputs)

if __name__ == '__main__':
    unittest.main()


#  This work is based on original code developed and copyrighted by TNO 2023.
#  Subsequent contributions are licensed to you by the developers of such code and are
#  made available to the Project under one or several contributor license agreements.
#
#  This work is licensed to you under the Apache License, Version 2.0.
#  You may obtain a copy of the license at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Contributors:
#      TNO         - Initial implementation
#  Manager:
#      TNO

import typing

from datetime import datetime
from dots_infrastructure.DataClasses import EsdlId, SimulaitonDataPoint
from dots_infrastructure.Logger import LOGGER
from influxdb import InfluxDBClient

from esdl import esdl

from dots_infrastructure.influxdb_connector import InfluxDBConnector



class InfluxDBMock(InfluxDBConnector):
    """A connector writes data to an InfluxDB database."""

    def __init__(self):
        super().__init__("test//test", "test", "test", "test", "test")
        self.data_points : typing.List[SimulaitonDataPoint] = []

    def connect(self) -> InfluxDBClient:
        LOGGER.info("Connecting to influxdb")

    def query(self, query):
        LOGGER.info("Query influxdb")

    def create_database(self):
        LOGGER.info("Create DB")

    def write(self, msgs):
        LOGGER.info("Writing to DB")

    def close(self):
        LOGGER.info("Close")

    def init_profile_output_data(
        self,
        simulation_id: str,
        model_id: str,
        esdl_type: str,
        esdl_objects: dict[EsdlId, esdl],
    ):
        self.simulation_id = simulation_id
        self.esdl_type = esdl_type
        self.model_id = model_id
        self.esdl_objects = esdl_objects

    def set_time_step_data_point(
        self, esdl_id: EsdlId, output_name: str, simulation_datetime: datetime, value: float
    ):
        self.data_points.append(SimulaitonDataPoint(output_name, simulation_datetime, value, esdl_id))

    def write_output(self):
        LOGGER.info("write output")

    def add_measurement(self, points, esdl_id, timestamp, fields):
        LOGGER.info("add measurement")



from datetime import datetime
import os
import helics as h
from typing import List

from dots_infrastructure.DataClasses import CalculationServiceOutput, PublicationDescription, SimulatorConfiguration
from dots_infrastructure.Logger import LOGGER

log_level_to_helics_log_level = {
    "debug" : h.HelicsLogLevel.DEBUG,
    "info" : h.HelicsLogLevel.TRACE,
    "warning" : h.HelicsLogLevel.WARNING,
    "error" : h.HelicsLogLevel.ERROR
}

def get_simulator_configuration_from_environment() -> SimulatorConfiguration:
    esdl_ids = os.getenv("esdl_ids", "e1b3dc89-cee8-4f8e-81ce-a0cb6726c17e;f006d594-0743-4de5-a589-a6c2350898da").split(";")
    esdl_type = os.getenv("esdl_type", "test-type")
    model_id = os.getenv("model_id", "test-id")
    broker_ip = os.getenv("broker_ip", "127.0.0.1")
    broker_port = int(os.getenv("broker_port", "30000"))
    start_time_str = str(os.getenv("start_time", "2024-06-10 09:51:13"))
    simulation_duration_in_seconds = int(os.getenv("simulation_duration_in_seconds", 86400))
    start_time_datetime = datetime.strptime(start_time_str, "%Y-%m-%d %H:%M:%S")
    calculation_services = os.getenv("calculation_services").split(";")
    simulation_id = str(os.getenv("simulation_id", "test-id"))
    influx_host = os.getenv("INFLUXDB_HOST")
    influx_port = os.getenv("INFLUXDB_PORT")
    influx_username = os.getenv("INFLUXDB_USER")
    influx_password = os.getenv("INFLUXDB_PASSWORD")
    influx_database_name = os.getenv("INFLUXDB_NAME")
    log_level = os.getenv("log_level", "INFO") 
    LOGGER.info(f"Using log level {log_level.upper()}")
    LOGGER.setLevel(log_level.upper())
    return SimulatorConfiguration(esdl_type, esdl_ids, model_id, broker_ip, broker_port,simulation_id, simulation_duration_in_seconds, start_time_datetime, influx_host, influx_port, influx_username, influx_password, influx_database_name, log_level_to_helics_log_level[log_level], calculation_services)

def generate_publications_from_value_descriptions(value_descriptions : List[PublicationDescription], simulator_configuration : SimulatorConfiguration) -> List[CalculationServiceOutput]:
    ret_val = []
    for value_description in value_descriptions:
        for esdl_id in simulator_configuration.esdl_ids:
            ret_val.append(CalculationServiceOutput(value_description.global_flag, value_description.esdl_type, value_description.output_name, esdl_id, value_description.data_type, value_description.output_unit))
    return ret_val

def get_single_param_with_name(param_dict : dict, name : str):
    for key in param_dict.keys():
        key_splitted = key.split("/")
        if any(name == key_part for key_part in key_splitted):
            return param_dict[key]
        
def clear_dictionary_values(dictionary_to_clear : dict):
    return dictionary_to_clear.fromkeys(dictionary_to_clear, None)

def dictionary_has_values_for_all_keys(dictionary : dict):
    return all([False if v == None else True for v in dictionary.values()])

def get_vector_param_with_name(param_dict : dict, name : str):
    return [value for key, value in param_dict.items() if any(name == key_part for key_part in key.split("/"))]



from pathlib import Path

from jinja2 import Environment, FileSystemLoader
from re import sub

from dots_infrastructure.code_gen.code_meta_data import CalculationServiceMetaData

TEMPLATE_DIR = Path(__file__).parent / "templates"
JINJA_ENV = Environment(loader=FileSystemLoader(TEMPLATE_DIR))

class CodeGenerator:

    helics_data_type_to_python_data_type = {
        "STRING" : "str",
        "DOUBLE" : "float",
        "INT" : "int",
        "COMPLEX" : "complex",
        "VECTOR" : "List",
        "COMPLEX_VECTOR" : "List[complex]",
        "BOOLEAN" : "bool",
        "TIME" : "int"
    }

    def camel_case(self, s):
        s = sub(r"(_|-| )+", " ", s).title().replace(" ", "")
        return ''.join([s[0].upper(), s[1:]])
    
    def get_python_name(self, s):
        return sub(r"(_|-| )+", " ", s).lower().replace(" ", "_")
    
    def get_base_class_name(self, class_name):
        return f'{class_name}Base'

    def render_template(self, template_path: Path, output_dir: Path, output_file: Path, **data):
        # jinja expects a string, representing a relative path with forward slashes
        template_path_str = str(template_path.relative_to(TEMPLATE_DIR)).replace("\\", "/")
        template = JINJA_ENV.get_template(template_path_str)
        JINJA_ENV.trim_blocks = True
        JINJA_ENV.lstrip_blocks = True

        output = template.render(**data)
        output_dir.mkdir(parents=True, exist_ok=True)
        with output_file.open(mode="w", encoding="utf-8") as output_file:
            output_file.write(output)

    def _extract_valid_python_datatype(self, data_type : str, name : str):
        if data_type in self.helics_data_type_to_python_data_type:
            return self.helics_data_type_to_python_data_type[data_type]
        else:
            raise ValueError(f"Unsupported helics data type: {data_type} for {name}, expected one of {", ".join(self.helics_data_type_to_python_data_type.keys())}")

    def render_calculation_service(self, template_path: Path, json_data : str, output_dir: Path):

        dataset_meta_data: CalculationServiceMetaData = CalculationServiceMetaData.schema().loads(json_data)

        class_name = self.camel_case(dataset_meta_data.name)
        base_class_name = self.get_base_class_name(class_name)
        file_name = self.get_python_name(dataset_meta_data.name)
        output_file = output_dir / f"{file_name}_base.py"

        for calculation in dataset_meta_data.calculations:
            calculation.calculation_function_name = calculation.name.replace(" ", "_").replace("-", "_").replace(".", "_")
            for input in calculation.inputs:
                input.python_name = self.get_python_name(input.name)
                self._extract_valid_python_datatype(input.data_type, input.name)
            for output in calculation.outputs:
                output.python_name = self.get_python_name(output.name)
                output.python_data_type = self._extract_valid_python_datatype(output.data_type, output.name)

        self.render_template(
            template_path=template_path,
            output_dir=output_dir,
            output_file=output_file,
            calculations=dataset_meta_data.calculations,
            name=base_class_name,
            esdl_type=dataset_meta_data.esdl_type
        )

    def render_output_dataclasses(self, template_path: Path, json_data : str, output_dir: Path):
        dataset_meta_data: CalculationServiceMetaData = CalculationServiceMetaData.schema().loads(json_data)

        file_name = self.get_python_name(dataset_meta_data.name)
        output_file = output_dir / f"{file_name}_dataclasses.py"

        for calculation in dataset_meta_data.calculations:
            calculation.calculation_output_class_name = f"{self.camel_case(calculation.name)}Output"
            for output in calculation.outputs:
                output.python_data_type = self._extract_valid_python_datatype(output.data_type, output.name)
                output.python_name = self.get_python_name(output.name)
                
        self.render_template(
            template_path=template_path,
            output_dir=output_dir,
            output_file=output_file,
            calculations=dataset_meta_data.calculations
        )

    def render_documentation(self, template_path: Path, json_data : str, output_dir: Path):
        dataset_meta_data: CalculationServiceMetaData = CalculationServiceMetaData.schema().loads(json_data)

        output_file = output_dir / f"{dataset_meta_data.name}.md"

        dataset_meta_data.relevant_links = [] if dataset_meta_data.relevant_links is None else dataset_meta_data.relevant_links

        self.render_template(
            template_path=template_path,
            output_dir=output_dir,
            output_file=output_file,
            calculations=dataset_meta_data.calculations,
            name=dataset_meta_data.name,
            esdl_type=dataset_meta_data.esdl_type,
            description=dataset_meta_data.description,
            relevant_links=dataset_meta_data.relevant_links
        )

    def code_gen(self, input : str, code_output_dir : str, documentation_ouput_dir : str):
        render_funcs = {
            "calculation_service": (self.render_calculation_service, code_output_dir),
            "cs_documentation": (self.render_documentation, documentation_ouput_dir),
            "cs_data_classes": (self.render_output_dataclasses, code_output_dir)
        }

        # render attribute classes
        for template_name, render_func_output_pair in render_funcs.items():
            for template_path in TEMPLATE_DIR.rglob(f"{template_name}.*.jinja"):
                render_func = render_func_output_pair[0]
                output_dir = Path(render_func_output_pair[1])
                output_path = output_dir 
                output_path.parent.mkdir(parents=True, exist_ok=True)
                print(f"Generating file in {output_path}")
                render_func(template_path=template_path, json_data=input, output_dir=output_path)






from dots_infrastructure.code_gen.code_gen import CodeGenerator

code_generator = CodeGenerator()
with open("test.json", "r") as input_file:
    input_data = input_file.read()

code_generator.code_gen(input=input_data, code_output_dir="src", documentation_ouput_dir="documentation")


import logging

LOG_LEVEL = 'INFO'
LOGGER = logging.getLogger('CS-Logger')
LOGGER.setLevel(LOG_LEVEL)

logging.basicConfig(
    format='[%(asctime)s] [%(threadName)s][%(filename)s:%(lineno)d][%(name)s-%(levelname)s]: %(message)s',
    level=LOG_LEVEL
)


from dataclasses import dataclass
from typing import List

from dataclasses_json import DataClassJsonMixin

@dataclass
class CalculationServiceInputData(DataClassJsonMixin):
    name : str
    esdl_type : str
    description : str
    unit : str
    data_type : str
    python_name : str | None = None


@dataclass
class CalculationServiceOutputData(DataClassJsonMixin):
    name : str
    description : str
    unit : str
    data_type : str
    python_data_type : str | None = None
    python_name : str | None = None


@dataclass
class Calculation(DataClassJsonMixin):
    name : str
    description : str
    time_period_in_seconds : int
    offset_in_seconds : int
    inputs : List[CalculationServiceInputData]
    outputs : List[CalculationServiceOutputData]
    calculation_function_name : str | None = None
    calculation_output_class_name : str | None = None


@dataclass
class RelevantLink(DataClassJsonMixin):
    name : str
    url : str
    description : str


@dataclass
class CalculationServiceMetaData(DataClassJsonMixin):
    name : str
    esdl_type : str
    description : str
    calculations : List[Calculation]
    relevant_links : List[RelevantLink] | None = None
    keywords : List[str] | None = None


import unittest

from dots_infrastructure.CalculationServiceHelperFunctions import get_single_param_with_name, get_vector_param_with_name

class TestLogicAddingCalculations(unittest.TestCase):

    def test_get_single_param_with_name(self):
        # Arrange
        param_dict = {
            "a/b/active_power": 1,
            "a/b/reactive_power": 2
        }
        name = "active_power"

        # Execute
        result = get_single_param_with_name(param_dict, name)
        expected_result = 1

        # Assert
        self.assertEqual(result, expected_result)

    def test_get_vector_param_with_name(self):
        # Arrange
        param_dict = {
            "a/a/active_power": 1,
            "a/b/active_power": 2,
            "a/c/reactive_power": 3
        }
        name = "active_power"

        # Execute
        result = get_vector_param_with_name(param_dict, name)
        expected_result = [1, 2]

        # Assert
        self.assertEqual(result, expected_result)

if __name__ == '__main__':
    unittest.main()


from concurrent.futures import ThreadPoolExecutor
import dataclasses
from datetime import timedelta
import math
import traceback
from typing import List
import helics as h
from esdl import esdl

from dots_infrastructure import Common
from dots_infrastructure.Constants import TimeRequestType
from dots_infrastructure.DataClasses import CalculationServiceInput, CalculationServiceOutput, HelicsCalculationInformation, HelicsMessageFederateInformation, PublicationDescription, RunningStatus, SubscriptionDescription, TimeStepInformation
from dots_infrastructure.EsdlHelper import EsdlHelper
from dots_infrastructure.Logger import LOGGER
from dots_infrastructure import CalculationServiceHelperFunctions
from dots_infrastructure.influxdb_connector import InfluxDBConnector

class HelicsFederateExecutor:

    def __init__(self):
        self.simulator_configuration = CalculationServiceHelperFunctions.get_simulator_configuration_from_environment()

    def init_default_federate_info(self):
        federate_info = h.helicsCreateFederateInfo()
        h.helicsFederateInfoSetBroker(federate_info, self.simulator_configuration.broker_ip)
        h.helicsFederateInfoSetBrokerPort(federate_info, self.simulator_configuration.broker_port)
        h.helicsFederateInfoSetCoreType(federate_info, h.HelicsCoreType.ZMQ)
        h.helicsFederateInfoSetIntegerProperty(federate_info, h.HelicsProperty.INT_LOG_LEVEL, h.HelicsLogLevel.NO_PRINT)
        return federate_info

    def init_calculation_service_federate_info(self, info : HelicsCalculationInformation):
        federate_info = self.init_default_federate_info()
        h.helicsFederateInfoSetTimeProperty(federate_info, h.HelicsProperty.TIME_PERIOD, info.federate_time_period)
        h.helicsFederateInfoSetTimeProperty(federate_info, h.HelicsProperty.TIME_DELTA, info.time_delta)
        h.helicsFederateInfoSetTimeProperty(federate_info, h.HelicsProperty.TIME_OFFSET, info.offset)
        h.helicsFederateInfoSetFlagOption(federate_info, h.HelicsFederateFlag.UNINTERRUPTIBLE, info.uninterruptible)
        h.helicsFederateInfoSetFlagOption(federate_info, h.HelicsFederateFlag.WAIT_FOR_CURRENT_TIME_UPDATE, info.wait_for_current_time_update)
        h.helicsFederateInfoSetFlagOption(federate_info, h.HelicsFlag.TERMINATE_ON_ERROR, info.terminate_on_error)
        return federate_info

class HelicsEsdlMessageFederateExecutor(HelicsFederateExecutor):
    def __init__(self, info : HelicsMessageFederateInformation):
        super().__init__()
        self.helics_message_federate_information = info
        self.message_federate = None
        self.message_enpoint = None

    def init_federate(self):
        federate_info = self.init_default_federate_info()
        self.message_federate = h.helicsCreateMessageFederate(f"{self.simulator_configuration.model_id}", federate_info)
        self.message_enpoint = h.helicsFederateRegisterEndpoint(self.message_federate, self.helics_message_federate_information.endpoint_name)

    def wait_for_esdl_file(self) -> EsdlHelper:
        h.helicsFederateEnterExecutingMode(self.message_federate)
        esdl_file_base64 = ""
        while h.helicsFederateRequestTime(self.message_federate, h.HELICS_TIME_MAXTIME) != h.HELICS_TIME_MAXTIME:
            LOGGER.debug("Fetching an esdl message string at time")
            esdl_file_base64_part = ""
            if h.helicsEndpointHasMessage(self.message_enpoint):
                esdl_file_base64_part_bytes = h.helicsMessageGetBytes(h.helicsEndpointGetMessage(self.message_enpoint))
                esdl_file_base64_part = esdl_file_base64_part_bytes.decode()
                LOGGER.debug(f"Received part of esdl file with: {len(esdl_file_base64_part)} characters")
            esdl_file_base64 += esdl_file_base64_part

        LOGGER.debug("Destroying esdl message federate")
        Common.destroy_federate(self.message_federate)
        esdl_helper = EsdlHelper(esdl_file_base64)

        return esdl_helper

class HelicsValueFederateExecutor(HelicsFederateExecutor):

    WAIT_FOR_INPUT_ITERATION_DURATION_SECONDS = 0.002

    def __init__(self, info : HelicsCalculationInformation):
        super().__init__()
        self.input_dict : dict[str, List[CalculationServiceInput]] = {}
        self.output_dict : dict[str, List[CalculationServiceOutput]] = {}
        self.all_inputs : List[CalculationServiceInput] = []
        self.helics_value_federate_info = info
        self.energy_system : esdl.EnergySystem = None
        self.value_federate : h.HelicsValueFederate = None
        self.running_status = RunningStatus()

    def init_outputs(self, pubs : List[PublicationDescription], value_federate : h.HelicsValueFederate):
        outputs = CalculationServiceHelperFunctions.generate_publications_from_value_descriptions(pubs, self.simulator_configuration)
        for output in outputs:
            key = f'{output.esdl_asset_type}/{output.output_name}/{output.output_esdl_id}'
            LOGGER.debug(f"[{h.helicsFederateGetName(self.value_federate)}] Registering publication with key: {key}")
            if output.global_flag:
                pub = h.helicsFederateRegisterGlobalPublication(value_federate, key, output.output_type, output.output_unit)
            else:
                pub = h.helicsFederateRegisterPublication(value_federate, key, output.output_type, output.output_unit)
            output.helics_publication = pub
            if output.output_esdl_id in self.output_dict:
                self.output_dict[output.output_esdl_id].append(output)
            else:
                self.output_dict[output.output_esdl_id] = [output]

    def init_inputs(self, subs : List[SubscriptionDescription], esdl_helper : EsdlHelper, value_federate : h.HelicsValueFederate):
        inputs : List[CalculationServiceInput] = []
        for esdl_id in self.simulator_configuration.esdl_ids:
            inputs_for_esdl_object = esdl_helper.get_connected_input_esdl_objects(esdl_id, self.simulator_configuration.calculation_services, subs)
            self.remove_duplicate_subscriptions_and_update_inputs(inputs, inputs_for_esdl_object)
            self.input_dict[esdl_id] = inputs_for_esdl_object

        for input in inputs:
            LOGGER.debug(f"[{self.value_federate.name}] Subscribing to publication with key: {input.helics_sub_key}")
            sub = h.helicsFederateRegisterSubscription(value_federate, input.helics_sub_key, input.input_unit)
            input.helics_input = sub

        self.all_inputs = inputs
        LOGGER.debug(f"Registered inputs: {inputs}")

    def remove_duplicate_subscriptions_and_update_inputs(self, inputs : List[CalculationServiceInput], inputs_for_esdl_object : List[CalculationServiceInput]):
        for i, new_input in enumerate(inputs_for_esdl_object):
            existing_input = next((input for input in inputs if input.helics_sub_key == new_input.helics_sub_key), None)
            if existing_input:
                inputs_for_esdl_object[i] = existing_input
            else:
                inputs.append(new_input)

    def init_federate(self, esdl_helper : EsdlHelper):
        federate_info = self.init_calculation_service_federate_info(self.helics_value_federate_info)
        self.value_federate = h.helicsCreateValueFederate(f"{self.simulator_configuration.model_id}/{self.helics_value_federate_info.calculation_name}", federate_info)
        self.init_inputs(self.helics_value_federate_info.inputs, esdl_helper, self.value_federate)
        self.init_outputs(self.helics_value_federate_info.outputs, self.value_federate)
        self.energy_system = esdl_helper.energy_system

    def get_helics_value(self, helics_sub : CalculationServiceInput):
        ret_val = None
        input_type = helics_sub.input_type
        sub = helics_sub.helics_input
        if h.helicsInputIsUpdated(sub):
            LOGGER.debug(f"[{h.helicsFederateGetName(self.value_federate)}] Getting value for subscription: {helics_sub.helics_sub_key} with type: {helics_sub.input_type} updated time: {h.helicsInputLastUpdateTime(sub)}")
            if input_type == h.HelicsDataType.BOOLEAN:
                ret_val = h.helicsInputGetBoolean(sub)
            elif input_type == h.HelicsDataType.COMPLEX_VECTOR:
                ret_val = h.helicsInputGetComplexVector(sub)
            elif input_type == h.HelicsDataType.DOUBLE:
                ret_val = h.helicsInputGetDouble(sub)
            elif input_type == h.HelicsDataType.COMPLEX:
                ret_val = h.helicsInputGetComplex(sub)
            elif input_type == h.HelicsDataType.INT:
                ret_val = h.helicsInputGetInteger(sub)
            elif input_type == h.HelicsDataType.JSON:
                ret_val = h.helicsInputGetString(sub)
            elif input_type == h.HelicsDataType.NAMED_POINT:
                ret_val = h.helicsInputGetNamedPoint(sub)
            elif input_type == h.HelicsDataType.STRING:
                ret_val = h.helicsInputGetString(sub)
            elif input_type == h.HelicsDataType.RAW:
                ret_val = h.helicsInputGetRawValue(sub)
            elif input_type == h.HelicsDataType.TIME:
                ret_val = h.helicsInputGetTime(sub)
            elif input_type == h.HelicsDataType.VECTOR:
                ret_val = h.helicsInputGetVector(sub)
            elif input_type == h.HelicsDataType.ANY:
                ret_val = h.helicsInputGetBytes(sub)
            else:
                raise ValueError("Unsupported Helics Data Type")
            LOGGER.debug(f"[{h.helicsFederateGetName(self.value_federate)}] Got value: {ret_val} from {helics_sub.helics_sub_key}")

        if ret_val == None:
            LOGGER.debug(f"[{h.helicsFederateGetName(self.value_federate)}] No new value for input: {helics_sub.helics_sub_key}")
        return ret_val

    def publish_helics_value(self, helics_output : CalculationServiceOutput, value):
        LOGGER.debug(f"[{h.helicsFederateGetName(self.value_federate)}] Publishing value: {value} for publication: {helics_output.helics_publication.name} with type: {helics_output.output_type.name}")
        pub = helics_output.helics_publication
        output_type = helics_output.output_type
        if output_type == h.HelicsDataType.BOOLEAN:
            h.helicsPublicationPublishBoolean(pub, value)
        elif output_type == h.HelicsDataType.COMPLEX_VECTOR:
            h.helicsPublicationPublishComplexVector(pub, value)
        elif output_type == h.HelicsDataType.DOUBLE:
            h.helicsPublicationPublishDouble(pub, value)
        elif output_type == h.HelicsDataType.COMPLEX:
            h.helicsPublicationPublishComplex(pub, value)
        elif output_type == h.HelicsDataType.INT:
            h.helicsPublicationPublishInteger(pub, value)
        elif output_type == h.HelicsDataType.JSON:
            h.helicsPublicationPublishString(pub, value)
        elif output_type == h.HelicsDataType.NAMED_POINT:
            h.helicsPublicationPublishNamedPoint(pub, value)
        elif output_type == h.HelicsDataType.STRING:
            h.helicsPublicationPublishString(pub, value)
        elif output_type == h.HelicsDataType.RAW:
            h.helicsPublicationPublishRaw(pub, value)
        elif output_type == h.HelicsDataType.TIME:
            h.helicsPublicationPublishTime(pub, value)
        elif output_type == h.HelicsDataType.VECTOR:
            h.helicsPublicationPublishVector(pub, value)
        elif output_type == h.HelicsDataType.ANY:
            h.helicsPublicationPublishBytes(pub, value)
        else:
            raise ValueError("Unsupported Helics Data Type")

    def finalize_simulation(self):
        Common.destroy_federate(self.value_federate)
        self.running_status.terminated = True

    def start_value_federate(self):
        self.enter_simulation_loop()
        self.finalize_simulation()

    def initialize_and_start_federate(self, esdl_helper : EsdlHelper):
        LOGGER.debug(f"[{self.simulator_configuration.model_id}/{self.helics_value_federate_info.calculation_name}] Initializing federate")
        self.init_federate(esdl_helper)
        LOGGER.debug(f"[{self.simulator_configuration.model_id}/{self.helics_value_federate_info.calculation_name}] Starting federate")
        self.start_value_federate()

    def _compute_time_step_number(self, time_of_timestep_to_compute):
        return int(math.floor(time_of_timestep_to_compute / self.helics_value_federate_info.time_period_in_seconds))

    def _init_calculation_params(self):
        ret_val = {}
        for esdl_id in self.simulator_configuration.esdl_ids:
            ret_val[esdl_id] = {} 
        for esdl_id in self.simulator_configuration.esdl_ids:
            if esdl_id in self.input_dict:
                inputs = self.input_dict[esdl_id]
                for helics_input in inputs:
                    ret_val[esdl_id][helics_input.helics_sub_key] = None
        return ret_val
    
    def _init_input_dict(self):
        input_dict = {}
        for input in self.all_inputs:
            input_dict[input.helics_sub_key] = None
        return input_dict

    def _publish_outputs(self, esdl_id, pub_values):
        if len(self.helics_value_federate_info.outputs) > 0:
            outputs = self.output_dict[esdl_id]
            for output in outputs:
                value_to_publish = pub_values[output.output_name]
                self.publish_helics_value(output, value_to_publish)

    def _gather_new_inputs(self, calculation_params, input_dict):
        for input in self.all_inputs:
            new_value = self.get_helics_value(input)
            if new_value != None:
                input_dict[input.helics_sub_key] = new_value

        for esdl_id in self.simulator_configuration.esdl_ids:
            if esdl_id in self.input_dict.keys():
                inputs = self.input_dict[esdl_id]
                for helics_input in inputs:
                    calculation_params[esdl_id][helics_input.helics_sub_key] = input_dict[helics_input.helics_sub_key]

    def _get_request_time(self, granted_time):
        requested_time = 0
        if self.helics_value_federate_info.time_request_type == TimeRequestType.PERIOD:
            if granted_time == 0 and self.helics_value_federate_info.offset > 0:
                requested_time = self.helics_value_federate_info.offset
            else:
                requested_time = granted_time + self.helics_value_federate_info.time_period_in_seconds
        if self.helics_value_federate_info.time_request_type == TimeRequestType.ON_INPUT:
            requested_time = h.HELICS_TIME_MAXTIME
        return requested_time

    def _gather_all_required_inputs(self, calculation_params : dict, granted_time : h.HelicsTime):
        LOGGER.debug(f"[{h.helicsFederateGetName(self.value_federate)}] Gathering all inputs")
        input_dict = self._init_input_dict()
        self._gather_new_inputs(calculation_params, input_dict)
        new_granted_time = granted_time
        
        while not CalculationServiceHelperFunctions.dictionary_has_values_for_all_keys(input_dict):
            LOGGER.debug(f"[{h.helicsFederateGetName(self.value_federate)}] requesting max time again to wait for new inputs")
            new_granted_time = h.helicsFederateRequestTime(self.value_federate, h.HELICS_TIME_MAXTIME)
            self._gather_new_inputs(calculation_params, input_dict)
        return new_granted_time

    def enter_simulation_loop(self):
        LOGGER.info(f"[{h.helicsFederateGetName(self.value_federate)}] Entering HELICS execution mode {self.helics_value_federate_info.calculation_name}")
        h.helicsFederateEnterExecutingMode(self.value_federate)
        LOGGER.info(f"[{h.helicsFederateGetName(self.value_federate)}] Entered HELICS execution mode {self.helics_value_federate_info.calculation_name}")

        total_interval = self.simulator_configuration.simulation_duration_in_seconds
        max_time_step_number = self._compute_time_step_number(total_interval)
        granted_time = 0
        granted_time = self.request_new_granted_time(granted_time)
        terminate_requested = False
        calculation_params = self._init_calculation_params()
        while granted_time <= total_interval and not terminate_requested:

            time_step_number = self._compute_time_step_number(granted_time)
            time_step_information = TimeStepInformation(time_step_number, max_time_step_number)
            granted_time = self._gather_all_required_inputs(calculation_params, granted_time)
            simulator_time = self.simulator_configuration.start_time + timedelta(seconds = granted_time)

            for esdl_id in self.simulator_configuration.esdl_ids:
                try:
                    if not terminate_requested:
                        LOGGER.info(f"[{h.helicsFederateGetName(self.value_federate)}] Executing calculation {self.helics_value_federate_info.calculation_name} for esdl_id {esdl_id} at time {granted_time}")
                        pub_values = self.helics_value_federate_info.calculation_function(calculation_params[esdl_id], simulator_time, time_step_information, esdl_id, self.energy_system)

                        if dataclasses.is_dataclass(pub_values):
                            pub_values = dataclasses.asdict(pub_values)

                        LOGGER.info(f"[{h.helicsFederateGetName(self.value_federate)}] Finished calculation {self.helics_value_federate_info.calculation_name} for esdl_id {esdl_id} at time {granted_time}")
                        self._publish_outputs(esdl_id, pub_values)
                        calculation_params[esdl_id] = CalculationServiceHelperFunctions.clear_dictionary_values(calculation_params[esdl_id])
                except Exception:
                    LOGGER.info(f"[{h.helicsFederateGetName(self.value_federate)}] Exception occurred for esdl_id {esdl_id} at time {granted_time} terminating simulation...")
                    traceback.print_exc()
                    terminate_requested = True
                    self.running_status.exception = True

            LOGGER.info(f"[{h.helicsFederateGetName(self.value_federate)}] Finished {granted_time} of {total_interval} and terminate requested {terminate_requested}")
            granted_time = self.request_new_granted_time(granted_time)

        LOGGER.info(f"[{h.helicsFederateGetName(self.value_federate)}] Finalizing federate at {granted_time} of {total_interval} and terminate requested {terminate_requested}")

    def request_new_granted_time(self, granted_time):
        time_to_request = self._get_request_time(granted_time)
        LOGGER.debug(f"[{h.helicsFederateGetName(self.value_federate)}] Requesting time: {time_to_request} for calculation {self.helics_value_federate_info.calculation_name}")
        granted_time = h.helicsFederateRequestTime(self.value_federate, time_to_request)
        LOGGER.debug(f"[{h.helicsFederateGetName(self.value_federate)}] Time granted: {granted_time} for calculation {self.helics_value_federate_info.calculation_name}")
        return granted_time

class HelicsSimulationExecutor:

    def __init__(self):
        self.simulator_configuration = CalculationServiceHelperFunctions.get_simulator_configuration_from_environment()
        self.calculations: List[HelicsValueFederateExecutor] = []
        self.energy_system = None
        self.influx_connector = InfluxDBConnector(self.simulator_configuration.influx_host, self.simulator_configuration.influx_port, self.simulator_configuration.influx_username, self.simulator_configuration.influx_password, self.simulator_configuration.influx_database_name)

    def add_calculation(self, info : HelicsCalculationInformation):
        if info.inputs == None:
            info.inputs = []
        if info.outputs == None:
            info.outputs = []
        info.federate_time_period = info.time_period_in_seconds
        if len(info.inputs) > 0:
            info.time_request_type = TimeRequestType.ON_INPUT
            info.federate_time_period = 0
        self.calculations.append(HelicsValueFederateExecutor(info))

    def _get_esdl_from_so(self):
        esdl_message_federate = HelicsEsdlMessageFederateExecutor(HelicsMessageFederateInformation('esdl'))
        esdl_message_federate.init_federate()
        esdl_helper = esdl_message_federate.wait_for_esdl_file()
        return esdl_helper

    def _init_influxdb(self, esdl_helper : EsdlHelper):
        esdl_objects = esdl_helper.esdl_object_mapping
        self.influx_connector.init_profile_output_data(self.simulator_configuration.simulation_id, self.simulator_configuration.model_id, self.simulator_configuration.esdl_type, esdl_objects)
        self.influx_connector.connect()

    def init_calculation_service(self, energy_system : esdl.EnergySystem):
        pass

    def init_simulation(self) -> esdl.EnergySystem:
        esdl_helper = self._get_esdl_from_so()
        self._init_influxdb(esdl_helper)
        self.init_calculation_service(esdl_helper.energy_system)
        return esdl_helper
    
    def _assert_that_periods_of_calculation_are_smaller_than_simulation_duration(self):
        simulation_duration = self.simulator_configuration.simulation_duration_in_seconds
        for calculation in self.calculations:
            if calculation.helics_value_federate_info.time_period_in_seconds > simulation_duration:
                raise RuntimeError(f"Calculation: {calculation.helics_value_federate_info.calculation_name} has a period > the simulatino duration therefore it will not execute")

    def start_simulation(self):
        self._assert_that_periods_of_calculation_are_smaller_than_simulation_duration()
        esdl_helper = self.init_simulation()
        self.exe = ThreadPoolExecutor(len(self.calculations))
        for calculation in self.calculations:
            self.exe.submit(calculation.initialize_and_start_federate, esdl_helper)

    def stop_simulation(self):
        self.exe.shutdown()
        LOGGER.debug(f"Writing data to influx for calculation service {self.simulator_configuration.model_id}")
        self.influx_connector.write_output()
        if any([calculation.running_status.exception for calculation in self.calculations]):
            failed_calulation = next((calculation for calculation in self.calculations if calculation.running_status.exception == True), None)
            raise RuntimeError(f"Calculation service had an exception calculation: {failed_calulation.helics_value_federate_info.calculation_name} failed")


from base64 import b64decode
from typing import List
from esdl.esdl_handler import EnergySystemHandler

from esdl import esdl
from esdl import EnergySystem
from dots_infrastructure.DataClasses import CalculationServiceInput, EsdlId, SubscriptionDescription
from dots_infrastructure.Logger import LOGGER

class EsdlHelper:

    def __init__(self, esdl_string_base64):
        self.energy_system = self._get_energy_system_from_base64_encoded_esdl_string(esdl_string_base64)
        self.non_connected_esdl_ids = []
        self.esdl_object_mapping : dict = self._get_esdl_id_object_mapping(self.energy_system, self.non_connected_esdl_ids)

    def _get_energy_system_from_base64_encoded_esdl_string(self, esdl_string_base64) -> EnergySystem:
        esdl_string = b64decode(esdl_string_base64 + b"==".decode("utf-8")).decode("utf-8")
        LOGGER.debug(f"Parsing esdl string of {len(esdl_string)} characters")
        esh = EnergySystemHandler()
        esh.load_from_string(esdl_string)
        return esh.get_energy_system()
    
    def _get_esdl_id_object_mapping(self, energy_system : EnergySystem, non_connected_esdl_ids : List) -> esdl:
        ret_val = {}
        for obj in energy_system.eAllContents():
            if hasattr(obj, "id"):
                ret_val[obj.id] = obj
                if not isinstance(obj, esdl.EnergyAsset):
                    non_connected_esdl_ids.append(obj.id)
        ret_val[energy_system.id] = energy_system
        return ret_val
    
    def extract_calculation_service_name(self, calculation_services: List[str], esdl_obj) -> str:
        esdl_obj_type_name = type(esdl_obj).__name__
        name = next(
            (
                calc_service_name
                for calc_service_name in calculation_services
                if calc_service_name == esdl_obj_type_name
            ),
            None,
        )
        
        return name
    
    def add_connected_esdl_object(self, subscriptions: List[SubscriptionDescription], calculation_services: List[str], input_descriptions : List[SubscriptionDescription], connected_asset: esdl, simulator_asset: esdl.EnergyAsset):
        calc_service_name = self.extract_calculation_service_name(calculation_services, connected_asset)
    
        if calc_service_name:
            input_descriptions = [input_description for input_description in input_descriptions if input_description.esdl_type == calc_service_name]
            for input_description in input_descriptions:
                new_input = CalculationServiceInput(input_description.esdl_type, input_description.input_name, connected_asset.id, input_description.input_unit, input_description.input_type, simulator_asset.id)
                new_input.helics_sub_key = f'{new_input.esdl_asset_type}/{new_input.input_name}/{new_input.input_esdl_id}'
                if new_input not in subscriptions:
                    subscriptions.append(new_input)
    
    def add_calc_services_from_ports_recursive(
        self,
        calculation_services: List[str],
        connected_input_esdl_objects: List[CalculationServiceInput],
        input_descriptions : List[SubscriptionDescription],
        model_esdl_asset: esdl.EnergyAsset,
        start_asset : esdl.EnergyAsset,
        visited_assets : List[str]
    ):
        if model_esdl_asset.id not in visited_assets:
            visited_assets.append(model_esdl_asset.id)
            for port in model_esdl_asset.port:
                for connected_port in port.connectedTo:
                    connected_asset = connected_port.eContainer()
                    self.add_connected_esdl_object(
                        connected_input_esdl_objects, calculation_services, input_descriptions, connected_asset, start_asset
                    )
                    if connected_asset.port != None and connected_asset.port != []:
                        self.add_calc_services_from_ports_recursive(calculation_services, connected_input_esdl_objects, input_descriptions, connected_asset, start_asset, visited_assets)

    def add_calc_services_from_building(self, calculation_services: List[str], connected_input_esdl_objects: List[CalculationServiceInput], input_descriptions : List[SubscriptionDescription], model_esdl_asset: esdl.EnergyAsset, building : esdl.Building):
        for esdl_entity in building.eAllContents():
            if isinstance(esdl_entity, esdl.EnergyAsset):
                self.add_connected_esdl_object(
                        connected_input_esdl_objects, calculation_services, input_descriptions, esdl_entity, model_esdl_asset
                    )

    def add_calc_services_from_ports(
        self,
        calculation_services: List[str],
        connected_input_esdl_objects: List[CalculationServiceInput],
        input_descriptions : List[SubscriptionDescription],
        model_esdl_asset: esdl.EnergyAsset
    ):
        visited_assets = []
        for port in model_esdl_asset.port:
            if isinstance(port, esdl.InPort):
                for connected_asset in port.connectedTo:
                    visited_assets.append(connected_asset.eContainer().id)
        if isinstance(model_esdl_asset.eContainer(), esdl.Building):
            self.add_calc_services_from_building(calculation_services, connected_input_esdl_objects, input_descriptions, model_esdl_asset, model_esdl_asset.eContainer())

        self.add_calc_services_from_ports_recursive(calculation_services, connected_input_esdl_objects, input_descriptions, model_esdl_asset, model_esdl_asset, visited_assets)

    def add_calc_services_from_non_connected_objects(
        self,
        calculation_services: List[str],
        connected_input_esdl_objects: List[SubscriptionDescription],
        input_descriptions : List[SubscriptionDescription],
        model_esdl_asset: esdl.EnergyAsset
    ):
        for esdl_id in self.non_connected_esdl_ids:
            self.add_connected_esdl_object(
                connected_input_esdl_objects, calculation_services, input_descriptions, self.esdl_object_mapping[esdl_id], model_esdl_asset
            )
        self.add_connected_esdl_object(connected_input_esdl_objects, calculation_services, input_descriptions, self.energy_system, model_esdl_asset)
    
    def add_calc_services_from_all_objects(
        self,
        calculation_services: List[str],
        connected_input_esdl_objects: List[SubscriptionDescription],
        input_descriptions : List[SubscriptionDescription],
        model_esdl_asset: esdl.EnergyAsset
    ):
        for esdl_obj in self.esdl_object_mapping.values():
            self.add_connected_esdl_object(
                connected_input_esdl_objects, calculation_services, input_descriptions, esdl_obj, model_esdl_asset
            )
    
    def get_connected_input_esdl_objects(
        self,
        esdl_id: EsdlId,
        calculation_services: List[str],
        input_descriptions : List[SubscriptionDescription]
    ) -> List[CalculationServiceInput]:
        model_esdl_obj = self.esdl_object_mapping[esdl_id]
    
        connected_input_esdl_objects: List[CalculationServiceInput] = []
        if isinstance(model_esdl_obj, esdl.EnergyAsset):
            self.add_calc_services_from_ports(
                calculation_services, connected_input_esdl_objects, input_descriptions, model_esdl_obj
            )
            self.add_calc_services_from_non_connected_objects(
                calculation_services, connected_input_esdl_objects, input_descriptions, model_esdl_obj
            )
        else:
            self.add_calc_services_from_all_objects(
                calculation_services, connected_input_esdl_objects, input_descriptions, model_esdl_obj
            )
        return connected_input_esdl_objects


from datetime import datetime
import unittest
from unittest.mock import MagicMock, call
import uuid

from esdl import Battery
from influxdb import InfluxDBClient
from dots_infrastructure.test_infra.InfluxDBClientMock import InfluxDBClientMock

from dots_infrastructure.influxdb_connector import InfluxDBConnector

class TestInfluxDBWriterLogic(unittest.TestCase):

    def setUp(self):
        self.influx_connector = InfluxDBConnector("test-host", "test-port", "test-user", "test-pwd", "test-db-name")
        self.influx_connector.connect = MagicMock(return_value=InfluxDBClientMock())

    def test_writing_to_influx_happens_in_chunks(self):
        # Arrange
        test_battery_id = str(uuid.uuid4())
        battery = Battery(name='battery_test', id=test_battery_id, capacity=5.0*3.6e6 ,
                              chargeEfficiency=1.0, dischargeEfficiency=1.0, fillLevel=0.3,
                              maxChargeRate=2.5, maxDischargeRate=2.5, powerFactor=1.0)

        self.influx_connector.init_profile_output_data("test-sim-id", "test-model-id", "Battery", {test_battery_id : battery})
        data_items = []
        for i in range(0,200100):
            self.influx_connector.set_time_step_data_point(test_battery_id, "test-output", datetime(2024,1,1), i)
            data_items.append({
                "measurement": "Battery",
                "tags": {
                    "simulation_id": "test-sim-id",
                    "model_id": "test-model-id",
                    "esdl_id": test_battery_id,
                    "esdl_name": 'battery_test',
                },
                "time": datetime(2024,1,1),
                "fields": {"test-output" : i},
            })

        # Execute
        self.influx_connector.write_output()

        # Assert
        calls = [call(data_items[0:100000], database="test-db-name", time_precision="s"), call(data_items[100000:200000], database="test-db-name", time_precision="s"), call(data_items[200000:200100], database="test-db-name", time_precision="s")]
        self.influx_connector.client.write_points.assert_has_calls(calls)


if __name__ == '__main__':
    unittest.main()


from pathlib import Path
import unittest
from unittest.mock import MagicMock, call

from dots_infrastructure.code_gen.code_gen import CodeGenerator
from dots_infrastructure.code_gen.code_meta_data import Calculation, CalculationServiceInputData, CalculationServiceOutputData, RelevantLink


class TestLogicAddingCalculations(unittest.TestCase):
    def setUp(self):
        self.code_generator = CodeGenerator()
        self.code_generator.render_template = MagicMock()
        self.test_template_dir = Path("test_template_path")
        self.test_output_dir = Path("test_output_dir")
        
    def test_exception_is_raised_when_invalid_data_type_is_provided(self):
        # Arrange
        with open("test_fault_code_gen_input.json", mode="r") as json_input:
            input_json = json_input.read()

        # Execute and Assert
        with self.assertRaises(ValueError):
            self.code_generator.render_calculation_service(self.test_template_dir, input_json, self.test_output_dir)

        with self.assertRaises(ValueError):
            self.code_generator.render_output_dataclasses(self.test_template_dir, input_json, self.test_output_dir)

    def test_when_valid_input_is_supplied_render_template_cs_is_called_correctly(self):
        # Arrange
        with open("test_valid_input.json", mode="r") as json_input:
            input_json = json_input.read()

        # Execute
        self.code_generator.render_calculation_service(self.test_template_dir, input_json, self.test_output_dir)

        # Assert
        expected_inputs = [
            CalculationServiceInputData("input1", "Weather", "input 1 description", "K", "STRING", "input1"),
        ]
        expected_outputs = [
            CalculationServiceOutputData("output1", "output 1 description", "W", "DOUBLE", python_data_type='float', python_name='output1')
        ]
        expected_calculations = [
            Calculation("test calculation", "test", 900, 0, expected_inputs, expected_outputs, calculation_function_name="test_calculation", calculation_output_class_name=None)
        ]
        calls = [
            call(template_path=self.test_template_dir, output_dir=self.test_output_dir, output_file=Path("test_output_dir/test_base.py"), calculations=expected_calculations, name="TestBase", esdl_type="PVInstallation"),
        ]
        self.code_generator.render_template.assert_has_calls(calls)

    def test_when_valid_input_is_supplied_render_template_dataclasses_is_called_correctly(self):
        # Arrange
        with open("test_valid_input.json", mode="r") as json_input:
            input_json = json_input.read()

        # Execute
        self.code_generator.render_output_dataclasses(self.test_template_dir, input_json, self.test_output_dir)

        # Assert
        expected_inputs = [
            CalculationServiceInputData("input1", "Weather", "input 1 description", "K", "STRING", None),
        ]
        expected_outputs = [
            CalculationServiceOutputData("output1", "output 1 description", "W", "DOUBLE", python_data_type='float', python_name='output1')
        ]
        expected_calculations = [
            Calculation("test calculation", "test", 900, 0, expected_inputs, expected_outputs, calculation_function_name=None, calculation_output_class_name='TestCalculationOutput')
        ]
        calls = [
            call(template_path=self.test_template_dir, output_dir=self.test_output_dir, output_file=Path("test_output_dir/test_dataclasses.py"), calculations=expected_calculations),
        ]
        self.code_generator.render_template.assert_has_calls(calls)

if __name__ == '__main__':
    unittest.main()


import base64
from dataclasses import dataclass
from datetime import datetime
from typing import List
import unittest
from unittest.mock import MagicMock, call, patch

from esdl import EnergySystem
import helics as h

from dots_infrastructure import CalculationServiceHelperFunctions, Common
from dots_infrastructure.Constants import TimeRequestType
from dots_infrastructure.DataClasses import CalculationServiceInput, CalculationServiceOutput, EsdlId, HelicsCalculationInformation, HelicsMessageFederateInformation, PublicationDescription, SimulatorConfiguration, SubscriptionDescription, TimeStepInformation
from dots_infrastructure.HelicsFederateHelpers import HelicsEsdlMessageFederateExecutor, HelicsValueFederateExecutor, HelicsSimulationExecutor
from dots_infrastructure.Logger import LOGGER
from dots_infrastructure.test_infra.HelicsMocks import HelicsEndpointMock, HelicsFederateMock

LOGGER.disabled = True

def simulator_environment_e_logic_test():
    return SimulatorConfiguration("LogicTest", ["f006d594-0743-4de5-a589-a6c2350898da"], "Mock-LogicTest", "127.0.0.1", 2000, "test-id", 5, datetime(2024,1,1), "test-host", "test-port", "test-username", "test-password", "test-database-name", h.HelicsLogLevel.DEBUG, ["PVInstallation", "EConnection"])

@dataclass
class TestDataClass:
    output1 : str
    output2 : int
    output3 : List

class CalculationServiceEConnection(HelicsSimulationExecutor):

    def __init__(self):
        super().__init__()

        subscriptions_values = [
            SubscriptionDescription("PVInstallation", "PV_Dispatch", "W", h.HelicsDataType.DOUBLE)
        ]

        e_connection_period_in_seconds = 60

        calculation_information = HelicsCalculationInformation(time_period_in_seconds=e_connection_period_in_seconds, 
                                                               time_request_type=TimeRequestType.PERIOD,
                                                               offset=0,
                                                               wait_for_current_time_update=False, 
                                                               uninterruptible=False, 
                                                               terminate_on_error=True, 
                                                               calculation_name="EConnectionDispatch", 
                                                               inputs=subscriptions_values, 
                                                               outputs=None, 
                                                               calculation_function=self.e_connection_dispatch)
        self.add_calculation(calculation_information)

        publication_values = [
            PublicationDescription(True, "EConnection", "Schedule", "W", h.HelicsDataType.VECTOR)
        ]

        e_connection_period_scedule_in_seconds = 120

        calculation_information_schedule = HelicsCalculationInformation(time_period_in_seconds=e_connection_period_scedule_in_seconds,
                                                                        time_request_type=TimeRequestType.PERIOD,
                                                                        offset=0, 
                                                                        wait_for_current_time_update=False, 
                                                                        uninterruptible=False, 
                                                                        terminate_on_error=True, 
                                                                        calculation_name="EConnectionSchedule", 
                                                                        inputs=None, 
                                                                        outputs=publication_values, 
                                                                        calculation_function=self.e_connection_da_schedule)
        self.add_calculation(calculation_information_schedule)

    def e_connection_dispatch(self, param_dict : dict, simulation_time : datetime, time_step_number : TimeStepInformation, esdl_id : EsdlId, energy_system : EnergySystem):

        return 0
    
    def e_connection_da_schedule(self, param_dict : dict, simulation_time : datetime, time_step_number : TimeStepInformation, esdl_id : EsdlId, energy_system : EnergySystem):
        return [1.0,2.0,3.0]

class TestLogicAddingCalculations(unittest.TestCase):
    def setUp(self):
        self.get_sim_config_from_env = CalculationServiceHelperFunctions.get_simulator_configuration_from_environment 
        CalculationServiceHelperFunctions.get_simulator_configuration_from_environment = simulator_environment_e_logic_test

    def tearDown(self):
        CalculationServiceHelperFunctions.get_simulator_configuration_from_environment = simulator_environment_e_logic_test
        
    def test_simulation_none_input_output_sets_empty_inputs_and_outputs(self):

        # Execute
        cs_econnection = CalculationServiceEConnection()

        # Assert
        self.assertEqual(len(cs_econnection.calculations), 2)
        self.assertEqual(len(cs_econnection.calculations[0].helics_value_federate_info.inputs), 1)
        self.assertEqual(len(cs_econnection.calculations[0].helics_value_federate_info.outputs), 0)
        self.assertEqual(len(cs_econnection.calculations[1].helics_value_federate_info.inputs), 0)
        self.assertEqual(len(cs_econnection.calculations[1].helics_value_federate_info.outputs), 1)

class TestLogicRunningSimulation(unittest.TestCase):

    def setUp(self):
        with open("test.esdl", mode="r") as esdl_file:
            self.encoded_base64_esdl = base64.b64encode(esdl_file.read().encode('utf-8')).decode('utf-8')

        self.federate_get_name = h.helicsFederateGetName
        self.get_sim_config_from_env = CalculationServiceHelperFunctions.get_simulator_configuration_from_environment 
        self.fed_eneter_executing_mode = h.helicsFederateEnterExecutingMode
        self.get_time_property = h.helicsFederateGetTimeProperty 
        self.request_time = h.helicsFederateRequestTime
        self.helics_endpoint_has_message = h.helicsEndpointHasMessage
        self.helics_message_get_bytes = h.helicsMessageGetBytes
        self.helics_endpoint_get_message = h.helicsEndpointGetMessage
        self.helics_create_message_federate = h.helicsCreateMessageFederate
        self.helics_federate_register_endpoint = h.helicsFederateRegisterEndpoint
        self.common_destroy_federate = Common.destroy_federate
        h.helicsFederateGetName = MagicMock(return_value = "LogicTest")
        CalculationServiceHelperFunctions.get_simulator_configuration_from_environment = simulator_environment_e_logic_test
        h.helicsFederateEnterExecutingMode = MagicMock()
        h.helicsFederateGetTimeProperty = MagicMock(return_value = 5)

        def helics_request_time(a, b):
            return b

        h.helicsFederateRequestTime = MagicMock(side_effect=helics_request_time)
        h.helicsEndpointHasMessage = MagicMock(return_value = True)
        h.helicsEndpointGetMessage = MagicMock(return_value = None)
        h.helicsCreateMessageFederate = MagicMock(return_value = HelicsFederateMock())
        h.helicsFederateRegisterEndpoint = MagicMock(return_value = HelicsEndpointMock())
        Common.destroy_federate = MagicMock()

    def tearDown(self):
        h.helicsFederateGetName = self.federate_get_name
        CalculationServiceHelperFunctions.get_simulator_configuration_from_environment = self.get_sim_config_from_env 
        h.helicsFederateEnterExecutingMode = self.fed_eneter_executing_mode
        h.helicsFederateGetTimeProperty = self.get_time_property
        h.helicsFederateRequestTime = self.request_time
        h.helicsEndpointHasMessage = self.helics_endpoint_has_message
        h.helicsMessageGetBytes = self.helics_message_get_bytes
        h.helicsCreateMessageFederate = self.helics_create_message_federate 
        h.helicsFederateRegisterEndpoint = self.helics_federate_register_endpoint
        h.helicsEndpointGetMessage = self.helics_endpoint_get_message
        Common.destroy_federate = self.common_destroy_federate

    def test_helics_simulation_loop_started_correctly(self):
        calculation_information_schedule = HelicsCalculationInformation(time_period_in_seconds=5,
                                                                        offset=0,
                                                                        wait_for_current_time_update=False, 
                                                                        uninterruptible=False, 
                                                                        terminate_on_error=True, 
                                                                        calculation_name="EConnectionSchedule", 
                                                                        inputs=[], 
                                                                        outputs=[], 
                                                                        calculation_function=MagicMock(return_value=5))
        self.federate_executor = HelicsValueFederateExecutor(calculation_information_schedule)

        # Execute
        self.federate_executor.enter_simulation_loop()

        # Assert
        h.helicsFederateEnterExecutingMode.assert_called_once()
        calculation_information_schedule.calculation_function.assert_called_once_with({}, datetime(2024, 1, 1, 0, 0, 5), TimeStepInformation(1,1), 'f006d594-0743-4de5-a589-a6c2350898da', None)

    def test_when_calculation_has_period_bigger_than_simulation_duration_then_exception_is_raised(self):
        # arrange
        calculation_information_schedule = HelicsCalculationInformation(time_period_in_seconds=10,
                                                                        offset=0,
                                                                        wait_for_current_time_update=False, 
                                                                        uninterruptible=False, 
                                                                        terminate_on_error=True, 
                                                                        calculation_name="EConnectionSchedule", 
                                                                        inputs=[], 
                                                                        outputs=[], 
                                                                        calculation_function=MagicMock(return_value=5),
                                                                        time_delta=0)
        
        executor = HelicsSimulationExecutor()
        executor.init_simulation = MagicMock()
        executor.add_calculation(calculation_information_schedule)
        executor.calculations[0].initialize_and_start_federate = MagicMock()

        # Execute and assert
        with self.assertRaises(RuntimeError):
            executor.start_simulation()
    
    def test_when_time_request_type_period_helicsFederateRequestTime_called_with_period(self):
        # arrange
        calculation_information_schedule = HelicsCalculationInformation(time_period_in_seconds=5,
                                                                        offset=0,
                                                                        wait_for_current_time_update=False, 
                                                                        uninterruptible=False, 
                                                                        terminate_on_error=True, 
                                                                        calculation_name="EConnectionSchedule", 
                                                                        inputs=[], 
                                                                        outputs=[], 
                                                                        calculation_function=MagicMock(return_value=5),
                                                                        time_delta=0)
        self.federate_executor = HelicsValueFederateExecutor(calculation_information_schedule)

        # Execute
        self.federate_executor.enter_simulation_loop()

        # Assert
        h.helicsFederateRequestTime.assert_has_calls([call(None, 5), call(None, 10)])

    def test_when_time_request_type_on_input_helicsFederateRequestTime_called_with_helics_max_time(self):
        # arrange
        calculation_information_schedule = HelicsCalculationInformation(time_period_in_seconds=60,
                                                                        offset=0,
                                                                        wait_for_current_time_update=False, 
                                                                        uninterruptible=False, 
                                                                        terminate_on_error=True, 
                                                                        calculation_name="EConnectionSchedule", 
                                                                        inputs=[], 
                                                                        outputs=[], 
                                                                        calculation_function=MagicMock(return_value=5),
                                                                        time_request_type=TimeRequestType.ON_INPUT,
                                                                        time_delta=5)
        self.federate_executor = HelicsValueFederateExecutor(calculation_information_schedule)

        # Execute
        self.federate_executor.enter_simulation_loop()

        # Assert
        h.helicsFederateRequestTime.assert_called_once_with(None, h.HELICS_TIME_MAXTIME)

    def test_when_time_request_type_period_and_has_offset_helicsFederateRequestTime_called_with_offset(self):

        calculation_information_schedule = HelicsCalculationInformation(time_period_in_seconds=2,
                                                                        offset=1,
                                                                        wait_for_current_time_update=False, 
                                                                        uninterruptible=False, 
                                                                        terminate_on_error=True, 
                                                                        calculation_name="EConnectionSchedule", 
                                                                        inputs=[], 
                                                                        outputs=[], 
                                                                        calculation_function=MagicMock(return_value=5),
                                                                        time_request_type=TimeRequestType.PERIOD,
                                                                        time_delta=0)
        self.federate_executor = HelicsValueFederateExecutor(calculation_information_schedule)

        # Execute
        self.federate_executor.enter_simulation_loop()

        # Assert
        h.helicsFederateRequestTime.assert_has_calls([call(None, 1), call(None, 3), call(None, 5), call(None, 7)])

    def test_calculation_is_not_executed_when_all_inputs_are_not_present(self):

        calculation_function = MagicMock()
        # arrange
        calculation_information_schedule = HelicsCalculationInformation(time_period_in_seconds=5,
                                                                        offset=0,
                                                                        wait_for_current_time_update=False, 
                                                                        uninterruptible=False, 
                                                                        terminate_on_error=True, 
                                                                        calculation_name="EConnectionSchedule", 
                                                                        inputs=[], 
                                                                        outputs=[], 
                                                                        calculation_function=calculation_function)

        self.federate_executor = HelicsValueFederateExecutor(calculation_information_schedule)
        inputs = [
            CalculationServiceInput("test-type", "test-input", "test-input-id", "W", h.HelicsDataType.DOUBLE, "test-id", "test-input-key"),
            CalculationServiceInput("test-type2", "test-input2", "test-input-id2", "W", h.HelicsDataType.DOUBLE, "test-id", "test-input-key2")
        ]
        self.federate_executor.input_dict["f006d594-0743-4de5-a589-a6c2350898da"] = inputs
        self.federate_executor.all_inputs = inputs

        self.return_value = 0

        def helics_value_side_effect(value):
            self.return_value += 1
            if self.return_value > 3:
                return self.return_value 
            return self.return_value  if value == inputs[0] else None

        self.federate_executor.get_helics_value = MagicMock(side_effect=helics_value_side_effect)

        # Execute
        self.federate_executor.enter_simulation_loop()

        # Assert
        self.assertEqual(self.return_value, 4)
        calculation_function.assert_called_once()

    def test_calculation_is_executed_when_all_inputs_are_present(self):
        calculation_function = MagicMock()
        # arrange
        calculation_information_schedule = HelicsCalculationInformation(time_period_in_seconds=5,
                                                                        offset=0,
                                                                        wait_for_current_time_update=False, 
                                                                        uninterruptible=False, 
                                                                        terminate_on_error=True, 
                                                                        calculation_name="EConnectionSchedule", 
                                                                        inputs=[], 
                                                                        outputs=[], 
                                                                        calculation_function=calculation_function)

        self.federate_executor = HelicsValueFederateExecutor(calculation_information_schedule)
        test_input_key = "test-input-key"
        test_input_key2 = "test-input-key2"
        inputs = [
            CalculationServiceInput("test-type", "test-input", "test-input-id", "W", h.HelicsDataType.DOUBLE, "test-id", "test-input-key"),
            CalculationServiceInput("test-type2", "test-input2", "test-input-id2", "W", h.HelicsDataType.DOUBLE, "test-id", "test-input-key2")
        ]
        self.federate_executor.input_dict["f006d594-0743-4de5-a589-a6c2350898da"] = inputs
        self.federate_executor.all_inputs = inputs

        self.federate_executor.get_helics_value = MagicMock(return_value=5)
        param_dict = {
            test_input_key : 5,
            test_input_key2 : 5
        }

        # Execute
        self.federate_executor.enter_simulation_loop()

        # Assert
        calculation_function.assert_called_once_with(param_dict, datetime(2024, 1, 1, 0, 0, 5), TimeStepInformation(1, 1), 'f006d594-0743-4de5-a589-a6c2350898da', None)

    def test_calculation_can_provide_dataclasses_as_output(self):
        calculation_function = MagicMock(return_value=TestDataClass(output1="test", output2=5, output3=[1, 2, 3]))
        # arrange
        calculation_information_schedule = HelicsCalculationInformation(time_period_in_seconds=5,
                                                                        offset=0,
                                                                        wait_for_current_time_update=False, 
                                                                        uninterruptible=False, 
                                                                        terminate_on_error=True, 
                                                                        calculation_name="EConnectionSchedule", 
                                                                        inputs=[], 
                                                                        outputs=[], 
                                                                        calculation_function=calculation_function)

        self.federate_executor = HelicsValueFederateExecutor(calculation_information_schedule)
        self.federate_executor._publish_outputs = MagicMock()

        # Execute
        self.federate_executor.enter_simulation_loop()

        # Assert
        expected_output_dict = {
            "output1": "test",
            "output2": 5,
            "output3": [1, 2, 3]
        }
        self.federate_executor._publish_outputs.assert_called_once_with('f006d594-0743-4de5-a589-a6c2350898da', expected_output_dict)

    def test_add_calculation_sets_correct_delta_and_period_values(self):
        calculation_function = MagicMock()
        # arrange
        calculation_information_schedule = HelicsCalculationInformation(time_period_in_seconds=5,
                                                                        offset=0,
                                                                        wait_for_current_time_update=False, 
                                                                        uninterruptible=False, 
                                                                        terminate_on_error=True, 
                                                                        calculation_name="EConnectionSchedule", 
                                                                        inputs=[CalculationServiceInput("test-type", "test-input", "test-input-id", "W", h.HelicsDataType.DOUBLE, "test-id", "test-input-key"),], 
                                                                        outputs=[], 
                                                                        calculation_function=calculation_function)
        
        calculation_information_dispatch = HelicsCalculationInformation(time_period_in_seconds=5,
                                                                        offset=0,
                                                                        wait_for_current_time_update=False, 
                                                                        uninterruptible=False, 
                                                                        terminate_on_error=True, 
                                                                        calculation_name="EConnectionDispatch", 
                                                                        inputs=[], 
                                                                        outputs=[CalculationServiceOutput(True, "test-type", "test-output", "test-output-id", h.HelicsDataType.DOUBLE, "W")], 
                                                                        calculation_function=calculation_function)

        simulation_executor = HelicsSimulationExecutor()

        # Execute
        simulation_executor.add_calculation(calculation_information_schedule)
        simulation_executor.add_calculation(calculation_information_dispatch)

        # Assert
        self.assertEqual(len(simulation_executor.calculations), 2)

        self.assertEqual(simulation_executor.calculations[0].helics_value_federate_info.time_request_type, TimeRequestType.ON_INPUT )
        self.assertEqual(simulation_executor.calculations[0].helics_value_federate_info.time_period_in_seconds, 5)
        self.assertEqual(simulation_executor.calculations[0].helics_value_federate_info.federate_time_period, 0)

        self.assertEqual(simulation_executor.calculations[1].helics_value_federate_info.time_request_type, TimeRequestType.PERIOD )
        self.assertEqual(simulation_executor.calculations[1].helics_value_federate_info.time_period_in_seconds, 5)
        self.assertEqual(simulation_executor.calculations[1].helics_value_federate_info.federate_time_period, 5)

    def test_given_waiting_for_esdl_when_full_esdl_is_received_esdl_helper_is_correctly_constructed(self):
        # arrange
        esdl_message_federate = HelicsEsdlMessageFederateExecutor(HelicsMessageFederateInformation('esdl'))
        esdl_message_federate.init_federate()

        self.i = 0
        def helics_request_time(a, b):
            self.i += 1
            return 5 if self.i == 1 else h.HELICS_TIME_MAXTIME

        h.helicsFederateRequestTime = MagicMock(side_effect=helics_request_time)
        h.helicsMessageGetBytes = MagicMock(return_value=self.encoded_base64_esdl.encode())

        # execute
        helper = esdl_message_federate.wait_for_esdl_file()

        self.assertIsNotNone(helper)

    def test_given_waiting_for_esdl_when_full_esdl_is_received_in_parts_esdl_helper_is_correctly_constructed(self):
        # arrange
        esdl_message_federate = HelicsEsdlMessageFederateExecutor(HelicsMessageFederateInformation('esdl'))
        esdl_message_federate.init_federate()

        self.i = 0
        def helics_request_time(a, b):
            self.i += 1
            return 5 if self.i <= 2 else h.HELICS_TIME_MAXTIME

        h.helicsFederateRequestTime = MagicMock(side_effect=helics_request_time)

        self.j = 0
        def helics_message_get_bytes(a):
            self.j += 1
            return self.encoded_base64_esdl[0:int(len(self.encoded_base64_esdl))].encode() if self.j == 1 else self.encoded_base64_esdl[int(len(self.encoded_base64_esdl)):].encode()

        h.helicsMessageGetBytes = MagicMock(side_effect=helics_message_get_bytes)

        # execute
        helper = esdl_message_federate.wait_for_esdl_file()

        self.assertIsNotNone(helper)

if __name__ == '__main__':
    unittest.main()


from dataclasses import dataclass
from datetime import datetime
from typing import List
import helics as h

from dots_infrastructure.Constants import TimeRequestType

EsdlId = str

@dataclass
class CalculationServiceInput:
    esdl_asset_type : str
    input_name : str
    input_esdl_id : str
    input_unit : str
    input_type : h.HelicsDataType
    simulator_esdl_id : str
    helics_sub_key : str = ""
    helics_input : h.HelicsInput = None

@dataclass 
class CalculationServiceOutput:
    global_flag : bool
    esdl_asset_type : str
    output_name : str
    output_esdl_id : str
    output_type : h.HelicsDataType
    output_unit : str
    helics_publication : h.HelicsPublication = None

@dataclass
class HelicsMessageFederateInformation:
    endpoint_name : str

@dataclass
class SubscriptionDescription:
    esdl_type : str
    input_name : str
    input_unit : str
    input_type : h.HelicsDataType

@dataclass
class PublicationDescription:
    global_flag : bool
    esdl_type : str
    output_name : str
    output_unit : str
    data_type : h.HelicsDataType

@dataclass
class HelicsCalculationInformation:
    time_period_in_seconds : float
    offset : int
    uninterruptible : bool
    wait_for_current_time_update : bool
    terminate_on_error : bool
    calculation_name : str
    inputs : List[SubscriptionDescription]
    outputs : List[PublicationDescription]
    calculation_function : any
    time_delta : float = 0
    federate_time_period = 0
    time_request_type : TimeRequestType = TimeRequestType.PERIOD

@dataclass
class SimulatorConfiguration:
    esdl_type : str
    esdl_ids : List[str]
    model_id : str
    broker_ip : str
    broker_port : int
    simulation_id : str
    simulation_duration_in_seconds : int
    start_time : datetime
    influx_host : str
    influx_port : str
    influx_username : str
    influx_password : str
    influx_database_name : str
    log_level : h.HelicsLogLevel
    calculation_services : List[str]

@dataclass
class SimulaitonDataPoint:
    output_name : str
    datapoint_time : datetime
    value : float
    esdl_id : EsdlId

@dataclass
class TimeStepInformation:
    current_time_step_number : int
    max_time_step_number : int

@dataclass
class RunningStatus:
    terminated : bool = False
    exception : bool = False


from unittest.mock import MagicMock

class InfluxDBClientMock:
    def __init__(self, *args, **kwargs):
        # Use MagicMock to mock all methods
        self.write_points = MagicMock(return_value=True)
        self.query = MagicMock(return_value=MagicMock(raw={"series": [{"name": "mock_series", "values": [[1, "mock_value"]], "columns": ["time", "value"]}]}))
        self.switch_database = MagicMock()
        self.create_database = MagicMock()
        self.drop_database = MagicMock()
        self.close = MagicMock()

        # Mock connection attributes
        self.host = kwargs.get("host", "localhost")
        self.port = kwargs.get("port", 8086)
        self.username = kwargs.get("username", "root")
        self.password = kwargs.get("password", "root")
        self.database = kwargs.get("database", "testdb")


#  This work is based on original code developed and copyrighted by TNO 2023.
#  Subsequent contributions are licensed to you by the developers of such code and are
#  made available to the Project under one or several contributor license agreements.
#
#  This work is licensed to you under the Apache License, Version 2.0.
#  You may obtain a copy of the license at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Contributors:
#      TNO         - Initial implementation
#  Manager:
#      TNO

import typing

from datetime import datetime
from dots_infrastructure.DataClasses import EsdlId
from dots_infrastructure.Logger import LOGGER
from influxdb import InfluxDBClient

from esdl import esdl



class InfluxDBConnector:
    """A connector writes data to an InfluxDB database."""

    def __init__(
        self,
        influx_host: str,
        influx_port: str,
        influx_user: str,
        influx_password: str,
        influx_database_name: str,
    ):
        self.influx_host: str = influx_host.split("//")[-1]
        self.influx_port: str = influx_port
        self.influx_database_name: str = influx_database_name
        self.influx_user: str = influx_user
        self.influx_password: str = influx_password

        LOGGER.debug("influx server: {}".format(self.influx_host))
        LOGGER.debug("influx port: {}".format(self.influx_port))
        LOGGER.debug("influx database: {}".format(self.influx_database_name))

        self.client: typing.Optional[InfluxDBClient] = None
        self.simulation_id: typing.Optional[str] = None
        self.model_id: typing.Optional[str] = None
        self.esdl_type: typing.Optional[str] = None
        self.esdl_objects: typing.Optional[dict[EsdlId, esdl]] = None

    def connect(self) -> InfluxDBClient:
        client = None
        try:
            LOGGER.debug("Connecting InfluxDBClient")
            client = InfluxDBClient(
                host=self.influx_host,
                port=self.influx_port,
                database=self.influx_database_name,
                username=self.influx_user,
                password=self.influx_password,
            )
            LOGGER.debug("InfluxDBClient ping: {}".format(client.ping()))
        except Exception as e:
            LOGGER.debug("Failed to connect to influx db: {}".format(e))
            if client:
                client.close()
        return client

    def query(self, query):
        if self.client is None:
            self.client = self.connect()

        return self.client.query(query)

    def create_database(self):
        if self.client is None:
            self.client = self.connect()
        self.client.create_database(self.influx_database_name)

    def write(self, msgs):
        if self.client is None:
            self.client = self.connect()

        # Send message to database.
        chunk_size = 100000
        for i in range(0, len(msgs), chunk_size):
            chunk = msgs[i:i + chunk_size]
            self.client.write_points(chunk, database=self.influx_database_name, time_precision="s")
        

    def close(self):
        if self.client:
            self.client.close()
        self.client = None

    def init_profile_output_data(
        self,
        simulation_id: str,
        model_id: str,
        esdl_type: str,
        esdl_objects: dict[EsdlId, esdl],
    ):
        self.simulation_id = simulation_id
        self.esdl_type = esdl_type
        self.model_id = model_id
        self.data_points : typing.List[dict] = []
        self.esdl_objects = esdl_objects

    def set_time_step_data_point(
        self, esdl_id: EsdlId, output_name: str, simulation_datetime: datetime, value: float
    ):
        fields = {
            output_name : value
        }
        self.data_points.append(self.add_measurement(esdl_id, simulation_datetime, fields))
        MAX_AMOUNT_OF_DB_POINTS = 100000
        if len(self.data_points) >= MAX_AMOUNT_OF_DB_POINTS:
            LOGGER.debug(f"Writing {len(self.data_points)} data points to InfluxDB")
            self.write_output()
            self.data_points.clear()

    def write_output(self):
        self.write(self.data_points)

    def add_measurement(self, esdl_id, timestamp, fields):
        try:
            if hasattr(self.esdl_objects[esdl_id], "name"):
                esdl_name = self.esdl_objects[esdl_id].name
            else:
                esdl_name = self.esdl_type
            item = {
                "measurement": f"{self.esdl_type}",
                "tags": {
                    "simulation_id": self.simulation_id,
                    "model_id": self.model_id,
                    "esdl_id": esdl_id,
                    "esdl_name": esdl_name,
                },
                "time": timestamp,
                "fields": fields,
            }
            return item
        except Exception as e:
            LOGGER.debug(f"Exception: {e} {e.args}")



import base64
from datetime import datetime, timedelta
from threading import Thread
import time
from typing import List
import unittest
import helics as h

from unittest.mock import MagicMock

from dots_infrastructure import CalculationServiceHelperFunctions
from dots_infrastructure.DataClasses import EsdlId, HelicsCalculationInformation, PublicationDescription, SimulatorConfiguration, SubscriptionDescription, TimeStepInformation
from dots_infrastructure.EsdlHelper import EsdlHelper
from dots_infrastructure.HelicsFederateHelpers import HelicsEsdlMessageFederateExecutor, HelicsSimulationExecutor
from dots_infrastructure.Logger import LOGGER
from dots_infrastructure.test_infra.InfluxDBMock import InfluxDBMock
from esdl.esdl import EnergySystem

BROKER_TEST_PORT = 23404
START_DATE_TIME = datetime(2024, 1, 1, 0, 0, 0)
SIMULATION_DURATION_IN_SECONDS = 120
CALCULATION_SERVICES = ["PVInstallation", "EConnection", "EnergyMarket", "Carriers"]
STR_INFLUX_TEST_PORT = "test-port"
INFLUX_USERNAME = "test-username"
INFLUX_PASSWORD = "test-password"
INFLUX_DB_NAME = "test-database-name"
INFLUX_HOST = "test-host"
SIMULATION_ID = "test-id"
BROKER_IP = "127.0.0.1"

MS_TO_BROKER_DISCONNECT = 60000

def simulator_environment_e_pv():
    return SimulatorConfiguration("PVInstallation", ['176af591-6d9d-4751-bb0f-fac7e99b1c3d','b8766109-5328-416f-9991-e81a5cada8a6'], "Mock-PV", BROKER_IP, BROKER_TEST_PORT, SIMULATION_ID, SIMULATION_DURATION_IN_SECONDS, START_DATE_TIME, INFLUX_HOST, STR_INFLUX_TEST_PORT, INFLUX_USERNAME, INFLUX_PASSWORD, INFLUX_DB_NAME, h.HelicsLogLevel.DEBUG, CALCULATION_SERVICES)

class CalculationServicePVDispatch(HelicsSimulationExecutor):

    def __init__(self):
        CalculationServiceHelperFunctions.get_simulator_configuration_from_environment = simulator_environment_e_pv
        super().__init__()
        self.influx_connector = InfluxDBMock()
        publictations_values = [
            PublicationDescription(True, "PVInstallation", "PV_Dispatch", "W", h.HelicsDataType.DOUBLE)
        ]
        subscriptions_values = []
        pv_installation_period_in_seconds = 30
        info = HelicsCalculationInformation(pv_installation_period_in_seconds, 0, False, False, True, "pvdispatch_calculation", subscriptions_values, publictations_values, self.pvdispatch_calculation)
        self.add_calculation(info)

        self.influx_connector = InfluxDBMock()
        publictations_values = []
        subscriptions_values = [
            SubscriptionDescription("EConnection", "EConnectionDispatch", "W", h.HelicsDataType.DOUBLE),
            SubscriptionDescription("EnergyMarket", "Price", "EUR", h.HelicsDataType.DOUBLE)
        ]
        pv_installation_period_in_seconds = 60
        info = HelicsCalculationInformation(pv_installation_period_in_seconds, 0, False, False, True, "process_econnection_dispatch", subscriptions_values, publictations_values, self.process_econnection_dispatch)
        self.add_calculation(info)


    def pvdispatch_calculation(self, param_dict : dict, simulation_time : datetime, time_step_number : TimeStepInformation, esdl_id : EsdlId, energy_system : EnergySystem):
        ret_val = {}
        ret_val["PV_Dispatch"] = time_step_number.current_time_step_number
        self.influx_connector.set_time_step_data_point(esdl_id, "PV_Dispatch", simulation_time, ret_val["PV_Dispatch"])
        return ret_val
    
    def process_econnection_dispatch(self, param_dict : dict, simulation_time : datetime, time_step_number : TimeStepInformation, esdl_id : EsdlId, energy_system : EnergySystem):
        ret_val = {}
        return ret_val

class CalculationServicePVDispatchMultipleOutputs(HelicsSimulationExecutor):
    def __init__(self):
        CalculationServiceHelperFunctions.get_simulator_configuration_from_environment = simulator_environment_e_pv
        super().__init__()
        self.influx_connector = InfluxDBMock()
        publictations_values = [
            PublicationDescription(True, "PVInstallation", "PV_Dispatch", "W", h.HelicsDataType.DOUBLE),
            PublicationDescription(True, "PVInstallation", "PV_Dispatch2", "W", h.HelicsDataType.DOUBLE)
        ]
        subscriptions_values = []
        pv_installation_period_in_seconds = 30
        info = HelicsCalculationInformation(pv_installation_period_in_seconds, 0, True, False, True, "pvdispatch_calculation", subscriptions_values, publictations_values, self.pvdispatch_calculation)
        self.add_calculation(info)


    def pvdispatch_calculation(self, param_dict : dict, simulation_time : datetime, time_step_number : TimeStepInformation, esdl_id : EsdlId, energy_system : EnergySystem):
        ret_val = {}
        ret_val["PV_Dispatch"] = time_step_number.current_time_step_number
        ret_val["PV_Dispatch2"] = time_step_number.current_time_step_number
        self.influx_connector.set_time_step_data_point(esdl_id, "PV_Dispatch", simulation_time, ret_val["PV_Dispatch"])
        self.influx_connector.set_time_step_data_point(esdl_id, "PV_Dispatch2", simulation_time, ret_val["PV_Dispatch2"])
        return ret_val

def simulator_environment_energy_market():
    return SimulatorConfiguration("EnergyMarket", ["b612fc89-a752-4a30-84bb-81ebffc56b50"], "Mock-MarketService", BROKER_IP, BROKER_TEST_PORT, SIMULATION_ID, SIMULATION_DURATION_IN_SECONDS, START_DATE_TIME, INFLUX_HOST, STR_INFLUX_TEST_PORT, INFLUX_USERNAME, INFLUX_PASSWORD, INFLUX_DB_NAME, h.HelicsLogLevel.DEBUG, CALCULATION_SERVICES)

class CalculationServiceMarketService(HelicsSimulationExecutor):

    def __init__(self):
        CalculationServiceHelperFunctions.get_simulator_configuration_from_environment = simulator_environment_energy_market
        super().__init__()
        self.influx_connector = InfluxDBMock()
        self.calculation_service_initialized = False

        publication_values = [
            PublicationDescription(True, "EnergyMarket", "Price", "EUR", h.HelicsDataType.DOUBLE),
        ]

        energy_market_period_in_seconds = 60

        calculation_information = HelicsCalculationInformation(energy_market_period_in_seconds, 0, False, False, True, "MarketPrice", None, publication_values, self.energy_market_price)

        self.add_calculation(calculation_information)

    def energy_market_price(self, param_dict : dict, simulation_time : datetime, time_step_number : TimeStepInformation, esdl_id : EsdlId, energy_system : EnergySystem):
        ret_val = {}
        ret_val["Price"] = time_step_number.current_time_step_number
        return ret_val

def simulator_environment_e_connection():
    return SimulatorConfiguration("EConnection", ["f006d594-0743-4de5-a589-a6c2350898da"], "Mock-Econnection", BROKER_IP, BROKER_TEST_PORT, SIMULATION_ID, SIMULATION_DURATION_IN_SECONDS, START_DATE_TIME, INFLUX_HOST, STR_INFLUX_TEST_PORT, INFLUX_USERNAME, INFLUX_PASSWORD, INFLUX_DB_NAME, h.HelicsLogLevel.DEBUG, CALCULATION_SERVICES)

class CalculationServiceEConnection(HelicsSimulationExecutor):

    def __init__(self):
        CalculationServiceHelperFunctions.get_simulator_configuration_from_environment = simulator_environment_e_connection
        super().__init__()
        self.influx_connector = InfluxDBMock()
        self.calculation_service_initialized = False

        subscriptions_values_dispatch = [
            SubscriptionDescription("PVInstallation", "PV_Dispatch", "W", h.HelicsDataType.DOUBLE)
        ]

        publication_values_dispatch = [
            PublicationDescription(True, "EConnection", "EConnectionDispatch", "W", h.HelicsDataType.DOUBLE)
        ]

        e_connection_period_in_seconds = 30

        calculation_information = HelicsCalculationInformation(e_connection_period_in_seconds, 0, False, False, True, "EConnectionDispatch", subscriptions_values_dispatch, publication_values_dispatch, self.e_connection_dispatch)
        self.add_calculation(calculation_information)

        subscriptions_values_schedule = [
            SubscriptionDescription("PVInstallation", "PV_Dispatch", "W", h.HelicsDataType.DOUBLE),
            SubscriptionDescription("EnergyMarket", "Price", "EUR", h.HelicsDataType.DOUBLE)
        ]

        e_connection_period_scedule_in_seconds = 60

        calculation_information_schedule = HelicsCalculationInformation(e_connection_period_scedule_in_seconds, 0, False, False, True, "EConnectionSchedule", subscriptions_values_schedule, None, self.e_connection_da_schedule)
        self.add_calculation(calculation_information_schedule)

    def init_calculation_service(self, energy_system: EnergySystem):
        self.calculation_service_initialized = True

    def e_connection_dispatch(self, param_dict : dict, simulation_time : datetime, time_step_number : TimeStepInformation, esdl_id : EsdlId, energy_system : EnergySystem):
        pv_dispatch = CalculationServiceHelperFunctions.get_single_param_with_name(param_dict, "PV_Dispatch")
        ret_val = {}
        time.sleep(1)
        ret_val["EConnectionDispatch"] = pv_dispatch
        self.influx_connector.set_time_step_data_point(esdl_id, "EConnectionDispatch", simulation_time, ret_val["EConnectionDispatch"])
        return ret_val
    
    def e_connection_da_schedule(self, param_dict : dict, simulation_time : datetime, time_step_number : TimeStepInformation, esdl_id : EsdlId, energy_system : EnergySystem):
        ret_val = {}
        pv_dispatch = CalculationServiceHelperFunctions.get_single_param_with_name(param_dict, "PV_Dispatch")
        price = CalculationServiceHelperFunctions.get_single_param_with_name(param_dict, "Price")
        ret_val["Schedule"] = [pv_dispatch * price , pv_dispatch * price, pv_dispatch * price]
        self.influx_connector.set_time_step_data_point(esdl_id, "DAScedule", simulation_time, ret_val["Schedule"])
        return ret_val

def simulator_environment_e_commonity():
    return SimulatorConfiguration("Carriers", ["02fafa20-a1bd-488e-a4db-f3c0ca7ff51a"], "Mock-Commonity", BROKER_IP, BROKER_TEST_PORT, SIMULATION_ID, SIMULATION_DURATION_IN_SECONDS, START_DATE_TIME, INFLUX_HOST, STR_INFLUX_TEST_PORT, INFLUX_USERNAME, INFLUX_PASSWORD, INFLUX_DB_NAME, h.HelicsLogLevel.DEBUG, CALCULATION_SERVICES)

class CalculationServiceElectricityCommodity(HelicsSimulationExecutor):
    def __init__(self):
        CalculationServiceHelperFunctions.get_simulator_configuration_from_environment = simulator_environment_e_commonity
        super().__init__()
        self.influx_connector = InfluxDBMock()

        subscriptions_values = [
            SubscriptionDescription("EConnection", "EConnectionDispatch", "W", h.HelicsDataType.DOUBLE)
        ]

        commodity_period_in_seconds = 60

        calculation_information = HelicsCalculationInformation(commodity_period_in_seconds, 0, False, False, True, "Loadflow", subscriptions_values, None, self.load_flow)
        self.add_calculation(calculation_information)

    def load_flow(self, param_dict : dict, simulation_time : datetime, time_step_number : TimeStepInformation, esdl_id : EsdlId, energy_system : EnergySystem):
        load = CalculationServiceHelperFunctions.get_single_param_with_name(param_dict, "EConnectionDispatch")
        self.influx_connector.set_time_step_data_point(esdl_id, "load_flow", simulation_time, load)

class CalculationServiceMultiplePvInputsEConnection(HelicsSimulationExecutor):
    def __init__(self):
        CalculationServiceHelperFunctions.get_simulator_configuration_from_environment = simulator_environment_e_connection
        super().__init__()
        self.influx_connector = InfluxDBMock()

        subscriptions_values = [
            SubscriptionDescription("PVInstallation", "PV_Dispatch", "W", h.HelicsDataType.DOUBLE),
            SubscriptionDescription("PVInstallation", "PV_Dispatch2", "W", h.HelicsDataType.DOUBLE)
        ]

        e_connection_period_in_seconds = 30

        calculation_information = HelicsCalculationInformation(e_connection_period_in_seconds, 0, False, False, True, "EConnectionDispatch", subscriptions_values, [], self.e_connection_dispatch)
        self.add_calculation(calculation_information)

    def e_connection_dispatch(self, param_dict : dict, simulation_time : datetime, time_step_number : TimeStepInformation, esdl_id : EsdlId, energy_system : EnergySystem):
        pv_dispatch = CalculationServiceHelperFunctions.get_single_param_with_name(param_dict, "PV_Dispatch")
        pv_dispatch2 = CalculationServiceHelperFunctions.get_single_param_with_name(param_dict, "PV_Dispatch2")
        ret_val = {}
        ret_val["EConnectionDispatch"] = pv_dispatch
        self.influx_connector.set_time_step_data_point(esdl_id, "PV_Dispatch", simulation_time, pv_dispatch)
        self.influx_connector.set_time_step_data_point(esdl_id, "PV_Dispatch2", simulation_time, pv_dispatch2)
        return ret_val

class CalculationServiceEConnectionException(HelicsSimulationExecutor):

    def __init__(self):
        CalculationServiceHelperFunctions.get_simulator_configuration_from_environment = simulator_environment_e_connection
        super().__init__()
        self.influx_connector = InfluxDBMock()

        subscriptions_values = [
            SubscriptionDescription("PVInstallation", "PV_Dispatch", "W", h.HelicsDataType.DOUBLE)
        ]

        publication_values = [
            PublicationDescription(True, "EConnection", "EConnectionDispatch", "W", h.HelicsDataType.DOUBLE)
        ]

        e_connection_period_in_seconds = 60

        calculation_information = HelicsCalculationInformation(e_connection_period_in_seconds, 0, False, False, True, "EConnectionDispatch", subscriptions_values, publication_values, self.e_connection_dispatch)
        self.add_calculation(calculation_information)

    def e_connection_dispatch(self, param_dict : dict, simulation_time : datetime, time_step_number : TimeStepInformation, esdl_id : EsdlId, energy_system : EnergySystem):
        raise Exception("Test-exception")

class CalculationServiceEConnectionControlLoop(HelicsSimulationExecutor):
    def __init__(self):
        CalculationServiceHelperFunctions.get_simulator_configuration_from_environment = simulator_environment_e_connection
        super().__init__()
        self.influx_connector = InfluxDBMock()

        subscriptions_values = [
            SubscriptionDescription("PVInstallation", "PV_Dispatch", "W", h.HelicsDataType.DOUBLE)
        ]

        publication_values_dispatch = [
            PublicationDescription(True, "EConnection", "EConnectionDispatch", "W", h.HelicsDataType.DOUBLE)
        ]

        e_connection_period_in_seconds = 30

        calculation_information = HelicsCalculationInformation(e_connection_period_in_seconds, 0, False, False, True, "EConnectionDispatch", subscriptions_values, publication_values_dispatch, self.e_connection_dispatch)
        self.add_calculation(calculation_information)

    def e_connection_dispatch(self, param_dict : dict, simulation_time : datetime, time_step_number : TimeStepInformation, esdl_id : EsdlId, energy_system : EnergySystem):
        pv_dispatch = CalculationServiceHelperFunctions.get_single_param_with_name(param_dict, "PV_Dispatch")
        ret_val = {}
        ret_val["EConnectionDispatch"] = pv_dispatch
        time.sleep(1)
        return ret_val

class CalculationServicePVDispatchControlLoop(HelicsSimulationExecutor):

    def __init__(self):
        CalculationServiceHelperFunctions.get_simulator_configuration_from_environment = simulator_environment_e_pv
        super().__init__()
        self.influx_connector = InfluxDBMock()
        publictations_values = [
            PublicationDescription(True, "PVInstallation", "PV_Dispatch", "W", h.HelicsDataType.DOUBLE)
        ]
        subscriptions_values = []
        pv_installation_period_in_seconds = 30
        info = HelicsCalculationInformation(pv_installation_period_in_seconds, 0, False, False, True, "pvdispatch_calculation", subscriptions_values, publictations_values, self.pvdispatch_calculation)
        self.add_calculation(info)

        self.influx_connector = InfluxDBMock()
        publictations_values = []
        subscriptions_values = [
            SubscriptionDescription("EConnection", "EConnectionDispatch", "W", h.HelicsDataType.DOUBLE)
        ]
        pv_installation_period_in_seconds = 30
        info = HelicsCalculationInformation(pv_installation_period_in_seconds, 0, False, False, True, "process_econnection_dispatch", subscriptions_values, publictations_values, self.process_econnection_dispatch)
        self.add_calculation(info)

    def init_calculation_service(self, energy_system : EnergySystem):
        self.next_pv_dispatch = 1.0

    def pvdispatch_calculation(self, param_dict : dict, simulation_time : datetime, time_step_number : TimeStepInformation, esdl_id : EsdlId, energy_system : EnergySystem):
        ret_val = {}
        ret_val["PV_Dispatch"] = self.next_pv_dispatch
        return ret_val
    
    def process_econnection_dispatch(self, param_dict : dict, simulation_time : datetime, time_step_number : TimeStepInformation, esdl_id : EsdlId, energy_system : EnergySystem):
        self.next_pv_dispatch = CalculationServiceHelperFunctions.get_single_param_with_name(param_dict, "EConnectionDispatch") + 1
        self.influx_connector.set_time_step_data_point(esdl_id, "PV_Dispatch", simulation_time, self.next_pv_dispatch)
        ret_val = {}
        return ret_val

class CalculationServiceEConnectionOffset(HelicsSimulationExecutor):
    
    def __init__(self):
        CalculationServiceHelperFunctions.get_simulator_configuration_from_environment = simulator_environment_e_connection
        super().__init__()
        self.influx_connector = InfluxDBMock()
        publictations_values = [
            PublicationDescription(True, "EConnection", "EConnection_DateTime", "Datetime", h.HelicsDataType.STRING)
        ]
        subscriptions_values = []
        econnection_installation_period_in_seconds = 60
        info = HelicsCalculationInformation(econnection_installation_period_in_seconds, 30, False, False, True, "econnection_calculation", subscriptions_values, publictations_values, self.econnection_calculation)
        self.add_calculation(info)

    def econnection_calculation(self, param_dict : dict, simulation_time : datetime, time_step_number : TimeStepInformation, esdl_id : EsdlId, energy_system : EnergySystem):
        ret_val = {}
        ret_val_datetime = simulation_time.strftime("%d-%m-%Y, %H:%M:%S")
        ret_val["EConnection_DateTime"] = ret_val_datetime
        return ret_val

class CalculationServiceCarriersOffset(HelicsSimulationExecutor):

    def __init__(self):
        CalculationServiceHelperFunctions.get_simulator_configuration_from_environment = simulator_environment_e_commonity
        super().__init__()
        self.influx_connector = InfluxDBMock()
        subscriptions_values = [
            SubscriptionDescription("EConnection", "EConnection_DateTime", "Datetime", h.HelicsDataType.STRING)
        ]
        carriers_period_in_seconds = 60
        info = HelicsCalculationInformation(carriers_period_in_seconds, 0, False, False, True, "carrier_calculation", subscriptions_values, [], self.carrier_calculation)
        self.add_calculation(info)

    def carrier_calculation(self, param_dict : dict, simulation_time : datetime, time_step_number : TimeStepInformation, esdl_id : EsdlId, energy_system : EnergySystem):
        received_datetime_str = CalculationServiceHelperFunctions.get_single_param_with_name(param_dict, "EConnection_DateTime")
        received_datetime = datetime.strptime(received_datetime_str, "%d-%m-%Y, %H:%M:%S")
        datetime_to_save = received_datetime + timedelta(seconds=10)
        self.influx_connector.set_time_step_data_point(esdl_id, "EConnection_DateTime", simulation_time, datetime_to_save.strftime("%d-%m-%Y, %H:%M:%S"))
        ret_val = {}
        return ret_val

class TestSimulation(unittest.TestCase):

    def start_helics_broker(self, federates):
        broker = h.helicsCreateBroker("zmq", "helics_broker_test", f"-f {federates} --loglevel=debug --timeout='60s' --globaltime --port {BROKER_TEST_PORT}")
        broker.wait_for_disconnect(MS_TO_BROKER_DISCONNECT)

    def setUp(self):
        with open("test.esdl", mode="r") as esdl_file:
            encoded_base64_esdl = base64.b64encode(esdl_file.read().encode('utf-8')).decode('utf-8')

        self.wait_for_esdl_file = HelicsEsdlMessageFederateExecutor.wait_for_esdl_file
        self.esdl_message_init_federate = HelicsEsdlMessageFederateExecutor.init_federate 
        self.calculationServiceHelperFunctions_get_simulator_configuration_from_environment = CalculationServiceHelperFunctions.get_simulator_configuration_from_environment
        HelicsEsdlMessageFederateExecutor.wait_for_esdl_file = MagicMock(return_value=EsdlHelper(encoded_base64_esdl))
        HelicsEsdlMessageFederateExecutor.init_federate = MagicMock()
        LOGGER.setLevel("DEBUG")

    def tearDown(self):
        HelicsEsdlMessageFederateExecutor.wait_for_esdl_file = self.wait_for_esdl_file 
        HelicsEsdlMessageFederateExecutor.init_federate = self.esdl_message_init_federate 
        CalculationServiceHelperFunctions.get_simulator_configuration_from_environment = self.calculationServiceHelperFunctions_get_simulator_configuration_from_environment

    def start_broker(self, n_federates):
        self.broker_thread = Thread(target = self.start_helics_broker, args = [ n_federates ])
        self.broker_thread.start()

    def stop_broker(self):
        self.broker_thread.join()

    def test_given_calculation_with_multiple_outputs_all_outputs_get_published_and_consumed(self):
        # Arrange 
        self.start_broker(2)

        expected_data_point_values_dispatch = []
        for i in range(1, 5):
            expected_data_point_values_dispatch.append(i)
            expected_data_point_values_dispatch.append(i)

        # Execute
        cs_econnection = CalculationServiceMultiplePvInputsEConnection()
        cs_dispatch = CalculationServicePVDispatchMultipleOutputs()

        cs_econnection.start_simulation()
        cs_dispatch.start_simulation()
        cs_econnection.stop_simulation()
        cs_dispatch.stop_simulation()
        self.stop_broker()

        # Assert
        actual_data_point_values_dispatch = [dp.value for dp in cs_econnection.influx_connector.data_points]
        self.assertListEqual(expected_data_point_values_dispatch, actual_data_point_values_dispatch)

    def test_given_exception_occurss_termination_bool_is_set_to_true(self):
        # Arrange 
        self.start_broker(2)
        e_connection_dispatch_period_in_seconds = 60
        pv_period = 30

        # Execute
        cs_econnection = CalculationServiceEConnectionException()
        cs_dispatch = CalculationServicePVDispatch()

        cs_econnection.start_simulation()
        cs_dispatch.start_simulation()
        
        with self.assertRaises(RuntimeError):
            cs_econnection.stop_simulation()
            cs_dispatch.stop_simulation()

        self.stop_broker()

        # Assert
        # No data should be generated as exception is generated right away
        self.assertEqual(len(cs_econnection.influx_connector.data_points), 0) 
        self.assertTrue(all([calculation.running_status.exception and calculation.running_status.terminated for calculation in cs_econnection.calculations]))
        
        # 2 pv panels produce data at most 3 times so
        self.assertLessEqual(len(cs_econnection.influx_connector.data_points), 2 * e_connection_dispatch_period_in_seconds / pv_period + 1)

    def test_given_a_set_of_calculation_services_then_simulation_executes_as_expected(self):
        # Arrange 
        self.start_broker(6)

        e_connection_dispatch_period_in_seconds = 30
        e_connection_period_scedule_in_seconds = 60
        pv_period = 30
        expected_data_point_values_dispatch = [i for i in range(1, 5)]
        expected_data_point_values_loadflow = expected_data_point_values_dispatch
        expected_data_point_values_schedule = [[i * 2.0 * i, i * 2.0 * i, i * 2.0 * i] for i in range(1, 3)]

        # Execute
        cs_econnection = CalculationServiceEConnection()
        cs_dispatch = CalculationServicePVDispatch()
        cs_market = CalculationServiceMarketService()
        cs_electricity_commodity = CalculationServiceElectricityCommodity()

        cs_econnection.start_simulation()
        cs_dispatch.start_simulation()
        cs_market.start_simulation()
        cs_electricity_commodity.start_simulation()
        cs_econnection.stop_simulation()
        cs_dispatch.stop_simulation()
        cs_market.stop_simulation()
        cs_electricity_commodity.stop_simulation()
        self.stop_broker()

        # Assert
        actual_data_point_values_dispatch = [dp.value for dp in cs_econnection.influx_connector.data_points if not isinstance(dp.value, List)]
        actual_data_point_values_schedule = [dp.value for dp in cs_econnection.influx_connector.data_points if isinstance(dp.value, List)]
        actual_data_point_values_loadflow = [dp.value for dp in cs_electricity_commodity.influx_connector.data_points]
        self.assertListEqual(expected_data_point_values_dispatch, actual_data_point_values_dispatch)
        self.assertListEqual(expected_data_point_values_schedule, actual_data_point_values_schedule)
        self.assertListEqual(expected_data_point_values_loadflow, actual_data_point_values_loadflow)
        self.assertEqual(len(cs_econnection.influx_connector.data_points), 
                         SIMULATION_DURATION_IN_SECONDS / e_connection_dispatch_period_in_seconds + 
                         SIMULATION_DURATION_IN_SECONDS / e_connection_period_scedule_in_seconds)
        self.assertEqual(len(cs_dispatch.influx_connector.data_points), SIMULATION_DURATION_IN_SECONDS / pv_period * 2)
        self.assertTrue(cs_econnection.calculation_service_initialized)

    def test_given_a_two_calculation_services_in_control_loop_then_data_is_exchanged_as_expected(self):
        # Arrange 
        self.start_broker(3)
        expected_data_point_values_dispatch = []
        for i in range(2, 6):
            expected_data_point_values_dispatch.append(i)
            expected_data_point_values_dispatch.append(i)

        # Execute
        cs_econnection = CalculationServiceEConnectionControlLoop()
        cs_dispatch = CalculationServicePVDispatchControlLoop()

        cs_econnection.start_simulation()
        cs_dispatch.start_simulation()
        cs_econnection.stop_simulation()
        cs_dispatch.stop_simulation()
        self.stop_broker()

        # Assert
        actual_data_point_values_dispatch = [dp.value for dp in cs_dispatch.influx_connector.data_points ]
        self.assertListEqual(expected_data_point_values_dispatch, actual_data_point_values_dispatch)

    def test_given_a_calculation_service_has_offset_then_first_execution_is_at_offset_time(self):
        # Arrange 
        self.start_broker(1)
        expected_data_point_values_dispatch = [datetime(2024, 1, 1, 0, 0, 40).strftime("%d-%m-%Y, %H:%M:%S"), datetime(2024, 1, 1, 0, 1, 40).strftime("%d-%m-%Y, %H:%M:%S")]

        # Execute
        cs_dispatch = CalculationServiceEConnectionOffset()
        cs_electricity_commodity = CalculationServiceCarriersOffset()
        cs_dispatch.start_simulation()
        cs_electricity_commodity.start_simulation()
        cs_dispatch.stop_simulation()
        cs_electricity_commodity.stop_simulation()
        self.stop_broker()

        # Assert
        actual_data_point_values_dispatch = [dp.value for dp in cs_electricity_commodity.influx_connector.data_points]
        self.assertListEqual(expected_data_point_values_dispatch, actual_data_point_values_dispatch)


if __name__ == '__main__':
    unittest.main()